import numpy as np

from pymoo.operators.repair.bounds_repair import BoundsRepair


def inverse_penality(x, p, xl, xu, alpha=None):
    assert len(p) == len(x)

    normv = np.linalg.norm(p - x)

    # violated boundaries
    idl = x < xl
    idr = xu < x

    # if nothing is out of bounds just return the value
    if not np.any(np.logical_or(idl, idr)):
        return x

    else:
        # lower bounds of Y
        diff = (p - x)
        diff[diff == 0] = 1e-32
        d = normv * np.max(np.maximum(idl * (xl - x) / diff, idr * (xu - x) / diff))

        # upper bounds on Y
        bounds = np.array([~idl * ((xl - x) / diff), ~idr * (xu - x) / diff])

        D = normv * np.min(bounds[bounds > 0])

        if alpha is None:
            alpha = (normv - d) / normv
            alpha += 1e-32

        r = np.random.random()
        Y = d * (1.0 + alpha * np.tan(r * np.arctan((D - d) / (alpha * d))))

        ret = x + (p - x) * Y / normv

        # for floating point error - in theory it will always be in bounds
        ret[ret < xl] = xl[ret < xl]
        ret[ret > xu] = xu[ret > xu]

        return ret


def inverse_penality_by_problem(problem, x, p, **kwargs):
    return inverse_penality(x, p, problem.xl, problem.xu, **kwargs)


class InversePenaltyOutOfBoundsRepair(BoundsRepair):

    def repair_out_of_bounds(self, problem, X, P=None, **kwargs):
        if P is None:
            raise Exception("For this out of bounds handling a parent solution in bounds needs to be provided.")
        assert len(X) == len(P)
        n = len(X)

        for k in range(n):
            X[k] = inverse_penality_by_problem(problem, X[k], P[k])

        return X


if __name__ == '__main__':

    # lower and upper bounds
    xl = np.zeros(2)
    xu = np.ones(2)

    # chosen parents
    # p = np.array([0.1, 1.0])
    p = np.array([0.5, 0.6])

    c = np.array([-0.1, 1.0])

    import matplotlib.pyplot as plt

    plt.scatter(p[0], p[1], color="green", label="Parent")
    plt.scatter(c[0], c[1], color="orange", label="Offspring")

    data = []
    for j in range(200):
        ret = inverse_penality(c, p, xl, xu, alpha=None)
        plt.scatter(ret[0], ret[1], facecolor="none", edgecolor="red", s=10, alpha=0.6)

    plt.ylim(0.0, 1.3)
    plt.xlim(-0.2, 1)
    plt.legend()
    plt.show()



import numpy as np

from pymoo.vendor.hv import HyperVolume


def hv(ref_point, F):
    hv = HyperVolume(ref_point)
    return hv.compute(F)


def calc_hvc_looped(ref_point, F, func=hv):
    hv = func(F, ref_point)

    hvi = []

    for k in range(len(F)):
        v = np.full(len(F), True)
        v[k] = False
        _hv = func(F[v], ref_point)
        hvi.append(hv - _hv)

    hvi = np.array(hvi)
    return hvi



import numpy as np

from pymoo.indicators.igd import IGD
from pymoo.termination.delta import DeltaToleranceTermination
from pymoo.util.normalization import normalize


class DesignSpaceTermination(DeltaToleranceTermination):

    def __init__(self, tol=0.005, **kwargs):
        super().__init__(tol, **kwargs)

    def _delta(self, prev, current):

        if prev.dtype == float and current.dtype == float:
            return IGD(current).do(prev)
        else:
            return np.mean([np.sum(e != prev, axis=1).max() / len(e) for e in current])

    def _data(self, algorithm):

        X = algorithm.opt.get("X")

        # do normalization if bounds are given
        problem = algorithm.problem
        if X.dtype != object and problem.has_bounds():
            X = normalize(X, xl=problem.xl, xu=problem.xu)

        return X



from pymoo.core.termination import Termination
from pymoo.util.sliding_window import SlidingWindow


class RobustTermination(Termination):

    def __init__(self,
                 termination,
                 period=30,
                 ) -> None:
        """

        Parameters
        ----------

        termination : Termination
            The termination criterion that shall become robust

        period : int
            The number of last generations to be considered for termination.

        """
        super().__init__()

        # create a collection in case number of max generation or evaluations is used
        self.termination = termination

        # the history calculated also in a sliding window
        self.history = SlidingWindow(period)

    def _update(self, algorithm):
        perc = self.termination.update(algorithm)
        self.history.append(perc)
        return min(self.history)



import numpy as np

from pymoo.core.indicator import Indicator
from pymoo.indicators.distance_indicator import derive_ideal_and_nadir_from_pf
from pymoo.util.misc import at_least_2d_array
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.vendor.hv import HyperVolume as _HyperVolume


class Hypervolume(Indicator):

    def __init__(self, ref_point=None, pf=None, nds=True, norm_ref_point=True, ideal=None, nadir=None, **kwargs):

        pf = at_least_2d_array(pf, extend_as="row")
        ideal, nadir = derive_ideal_and_nadir_from_pf(pf, ideal=ideal, nadir=nadir)

        super().__init__(ideal=ideal, nadir=nadir, **kwargs)
        # self.normalization = ZeroToOneNormalization(ideal, nadir)

        # whether the input should be checked for domination or not
        self.nds = nds

        # the reference point that shall be used - either derived from pf or provided
        ref_point = ref_point
        if ref_point is None:
            if pf is not None:
                ref_point = pf.max(axis=0)

        # we also have to normalize the reference point to have the same scales
        if norm_ref_point:
            ref_point = self.normalization.forward(ref_point)

        self.ref_point = ref_point
        assert self.ref_point is not None, "For Hypervolume a reference point needs to be provided!"

    def _do(self, F):
        if self.nds:
            non_dom = NonDominatedSorting().do(F, only_non_dominated_front=True)
            F = np.copy(F[non_dom, :])

        # calculate the hypervolume using a vendor library
        hv = _HyperVolume(self.ref_point)
        val = hv.compute(F)
        return val


class HV(Hypervolume):
    pass


def hvc_looped(ref_point, F, func):
    hv = func(ref_point, F)

    hvc = []

    for k in range(len(F)):
        v = np.full(len(F), True)
        v[k] = False
        _hv = func(ref_point, F[v])
        hvc.append(hv - _hv)

    hvc = np.array(hvc)
    return hvc



import numpy as np


def project_onto(y, p, n):
    """

    Project a point onto a plane.

    Parameters
    ----------
    y : The point that should be project
    p : A point which lies on the plane
    n : The normal vector of the plane

    Returns
    -------

    proj_y : The projection onto the plane

    """

    is_1d = False
    if len(y.shape) == 1:
        is_1d = True
        y = np.atleast_2d(y)

    # make sure the plane vector is normalized
    n = n / np.linalg.norm(n)

    proj_y = y - n[None, :] * ((y - p) @ n)[:, None]

    if is_1d:
        proj_y = proj_y[0]

    return proj_y


def project_onto_sum_equals_one_plane(y):
    n_dim = y.shape[-1]
    return project_onto(y, np.eye(n_dim)[0], np.ones(n_dim))


def project_onto_sum_equals_zero_plane(y):
    n_dim = y.shape[-1]
    return project_onto(y, np.zeros(n_dim), np.ones(n_dim))


def project_onto_unit_simplex(X):
    n_points, n_dim = X.shape

    # a boolean mask of violating values - less than 0
    I = X < 0

    # get all the points which are out of bounds and need to be fixed
    out_of_unit_simplex = np.where(I.sum(axis=1) > 0)[0]

    # now check for each point if it is still in bound
    for j in out_of_unit_simplex:

        # indices where the last point was already out of bounds
        subspace = np.logical_not(I[j])

        # project the bounds back onto the simplex in the subspace
        proj = np.zeros(n_dim)
        proj[subspace] = project_onto_sum_equals_one_plane(X[j][subspace])

        # set the result to the corresponding value
        X[j] = proj

        test1 = matrix_project_onto_sum_equals_one_plane(X[j][subspace])
        test2 = project_onto_sum_equals_one_plane(X[j][subspace])

        if not np.allclose(test1, test2):
            print("test")

        if np.any(X[j] < 0):
            print("test")


def project_onto_unit_simplex_recursive(X):
    # get all the points which are out of bounds and need to be fixed
    out_of_unit_simplex = np.where(np.any(X < 0, axis=1))[0]

    # now check for each point if it is still in bound
    for j in out_of_unit_simplex:

        while True:
            X[j, X[j] < 0] = 0

            # indices where the last point was already out of bounds
            subspace = X[j] > 0

            # project the bounds back onto the simplex in the subspace
            X[j, subspace] = project_onto_sum_equals_one_plane(X[j][subspace])

            if np.all(X[j] >= 0):
                break

    return X


def matrix_project_onto_sum_equals_one_plane(next):
    n_dim = len(next)
    P, S = np.eye(n_dim)[0], next

    # create for each subspace dimension a point on the hyperplane
    points = P + np.eye(n_dim)[1:]
    points /= points.sum(axis=1)[:, None]

    v = points - P
    s = S - P

    # solve a system of linear equations to project the point
    A = np.zeros((n_dim - 1, n_dim - 1))
    for i in range(n_dim - 1):
        for j in range(n_dim - 1):
            A[i, j] = np.dot(v[i], v[j])

    b = np.zeros(n_dim - 1)
    for i in range(n_dim - 1):
        b[i] = np.dot(s, v[i])

    x = np.linalg.solve(A, b)

    # finally calculate the projection onto the plane
    proj = (P + x @ v)

    return proj



from copy import deepcopy

import numpy as np

from pymoo.core.algorithm import Algorithm
from pymoo.core.evaluator import Evaluator
from pymoo.core.individual import Individual
from pymoo.core.meta import Meta
from pymoo.core.population import Population
from pymoo.core.problem import Problem


class AttachConfigEvaluator(Meta, Evaluator):

    def __init__(self, wrapped, config):
        super().__init__(wrapped)
        self.config = config

    def eval(self, problem: Problem, pop: Population, **kwargs):
        pop = super().eval(problem, pop, **kwargs)
        pop.apply(lambda ind: ind.set("config", self.config))


def copy_to_dict(src, dest):
    dest.clear()
    dest.update(**src)


class AdaptiveConstraintHandling(Meta, Algorithm):

    def __init__(self, algorithm):
        super().__init__(algorithm)

        self.config = Individual.default_config()
        self.config["cache"] = False

        self.default_config = deepcopy(self.config)
        self.adapted_config = deepcopy(self.config)

    def _setup(self, _, **kwargs):
        self.evaluator = AttachConfigEvaluator(self.evaluator, self.config)

    def _adapt(self, config, infills=None, **kwargs):
        pass

    def _initialize_advance(self, infills=None, **kwargs):
        copy_to_dict(self.adapted_config, self.config)
        super()._initialize_advance(infills=infills, **kwargs)
        copy_to_dict(self.default_config, self.config)

    def _advance(self, infills=None, **kwargs):
        copy_to_dict(self.adapted_config, self.config)
        super()._advance(infills=infills, **kwargs)
        copy_to_dict(self.default_config, self.config)

        self._adapt(self.adapted_config, infills=infills, **kwargs)

    def _infill(self):
        copy_to_dict(self.adapted_config, self.config)
        pop = super()._infill()
        copy_to_dict(self.default_config, self.config)
        return pop



import numpy as np

from pymoo.core.mutation import Mutation
from pymoo.operators.crossover.ox import random_sequence


def inversion_mutation(y, seq, inplace=True):
    y = y if inplace else np.copy(y)

    if seq is None:
        seq = random_sequence(len(y))
    start, end = seq

    y[start:end + 1] = np.flip(y[start:end + 1])
    return y


class InversionMutation(Mutation):

    def __init__(self, prob=1.0):
        """

        This mutation is applied to permutations. It randomly selects a segment of a chromosome and reverse its order.
        For instance, for the permutation `[1, 2, 3, 4, 5]` the segment can be `[2, 3, 4]` which results in `[1, 4, 3, 2, 5]`.

        Parameters
        ----------
        prob : float
            Probability to apply the mutation to the individual
            
        """
        super().__init__()
        self.prob = prob

    def _do(self, problem, X, **kwargs):
        Y = X.copy()
        for i, y in enumerate(X):
            if np.random.random() < self.prob:
                seq = random_sequence(len(y))
                Y[i] = inversion_mutation(y, seq, inplace=True)

        return Y



import numpy as np

from pymoo.visualization.util import parse_bounds, normalize
from pymoo.docs import parse_doc_string
from pymoo.core.plot import Plot
from pymoo.util.misc import set_if_none_from_tuples


class Heatmap(Plot):

    def __init__(self,
                 cmap="Blues",
                 order_by_objectives=False,
                 reverse=True,
                 solution_labels=True,
                 **kwargs):
        """

        Heatmap

        Parameters
        ----------------

        cmap : str
            The color map to be used.

        order_by_objectives : int or list
            Whether the result should be ordered by an objective. If false no order.
            Otherwise, either supply just the objective or a list. (it is lexicographically sorted).

        reverse : bool
            If true large values are white and small values the corresponding color. Otherwise, the other way around.

        solution_labels : bool or list
            If False no labels are plotted in the y axis. If true just the corresponding index. Otherwise the label provided.

        bounds : {bounds}

        labels : {labels}

        Other Parameters
        ----------------

        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}
        cmap : {cmap}


        """

        super().__init__(cmap=cmap, reverse=reverse, **kwargs)
        self.order_by_objectives = order_by_objectives
        self.solution_labels = solution_labels

        # set default style
        set_if_none_from_tuples(self.axis_style, ("interpolation", "nearest"), ("vmin", 0), ("vmax", 1))

    def _do(self):

        if len(self.to_plot) != 1:
            raise Exception("Only one element can be added to a heatmap.")

        # initial a figure with a single plot
        self.init_figure()

        # normalize the input
        bounds = parse_bounds(self.bounds, self.n_dim)
        to_plot_norm = normalize(self.to_plot, bounds, reverse=self.reverse)
        (F, kwargs) = to_plot_norm[0]

        # dot the sorting if required
        if self.order_by_objectives is not None and self.order_by_objectives is not False:

            if isinstance(self.order_by_objectives, list) and len(self.order_by_objectives) == self.n_dim:
                L = self.order_by_objectives
            elif isinstance(self.order_by_objectives, int):
                L = [i for i in range(F.shape[1]) if i != self.order_by_objectives]
                L.insert(0, self.order_by_objectives)
            else:
                L = range(self.n_dim)

            _F = [F[:, j] for j in L]
            I = np.lexsort(_F[::-1])
        else:
            I = np.arange(len(F))

        # plot the data
        self.ax.imshow(F[I], cmap=self.cmap, **self.axis_style)

        # set the x ticks and labels
        self.ax.set_xticks(np.arange(self.n_dim))
        self.ax.set_xticklabels(self.get_labels())

        # no solution labels should be used
        if self.solution_labels is None:
            pass

        # in case just true just use a number for each solution
        elif isinstance(self.solution_labels, bool) and self.solution_labels:
            self.solution_labels = np.arange(len(F)) + 1

        # otherwise use directly the label provided
        else:
            if len(self.solution_labels) != len(F):
                raise Exception(
                    "The labels provided for each solution must be equal to the number of solutions being plotted.")

        if self.solution_labels is None:
            self.ax.set_yticks([])
            self.ax.set_yticklabels([])

        else:

            # for ordered by objective apply it to labels
            self.solution_labels = [self.solution_labels[i] for i in I]

            self.ax.set_yticks(np.arange(len(F)))
            self.ax.set_yticklabels(self.solution_labels)


parse_doc_string(Heatmap.__init__)



import matplotlib.pyplot as plt
import numpy as np

from pymoo.core.problem import ElementwiseProblem


class FlowshopScheduling(ElementwiseProblem):

    def __init__(self, processing_times, **kwargs):
        """
        Flowshop scheduling problem.

        Parameters
        ----------
        processing_times : numpy.array
            Matrix where processing_time[i][j] is the processing time for job j on machine i.

        """

        n_machines, n_jobs = processing_times.shape
        self.records = processing_times

        super(FlowshopScheduling, self).__init__(
            n_var=n_jobs,
            n_obj=1,
            xl=0,
            xu=n_machines,
            vtype=int,
            **kwargs
        )

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = self.makespan(x)

    def makespan(self, x):
        machine_times = self.get_machine_times(x)
        # The makespan is the difference between the starting time of the first job and the latest finish time of any
        # job. Minimizing the makespan amounts to minimizing the total time it takes to process all jobs from start
        # to finish.
        makespan = machine_times[-1][-1] + self.records[-1][x[-1]]  # finish time of the last job
        return makespan

    def get_machine_times(self, x):
        n_machines, n_jobs = self.records.shape

        # A 2d array to store the starting time for each job on each machine
        # machine_times[i][j] --> starting time for the j-th job on machine i
        machine_times = [[] for _ in range(n_machines)]

        # Assign the initial job to the machines
        machine_times[0].append(0)
        for i in range(1, n_machines):
            # Start the next job when the previous one is finished
            machine_times[i].append(
                machine_times[i - 1][0] + self.records[i - 1][x[0]]
            )

        # Assign the remaining jobs
        for j in range(1, n_jobs):
            # For the first machine, we can put a job when the previous one is finished
            machine_times[0].append(
                machine_times[0][j - 1] + self.records[0][x[j - 1]]
            )

            # For the remaining machines, the starting time of the current job j is the max of the following two times:
            # 1. The finish time of the previous job on the current machine
            # 2. The finish time of the current job on the previous machine
            for i in range(1, n_machines):
                machine_times[i].append(
                    max(
                        machine_times[i][j - 1] + self.records[i][x[j - 1]],  # 1
                        machine_times[i - 1][j] + self.records[i - 1][x[j]]  # 2
                    )
                )
        return machine_times


def create_random_flowshop_problem(n_machines, n_jobs, seed=None):
    if seed is not None:
        np.random.seed(seed)
    T = np.random.random((n_machines, n_jobs)) * 50 + 50
    return FlowshopScheduling(T)


def visualize(problem, x, path=None, label=True):
    with plt.style.context('ggplot'):
        n_machines, n_jobs = problem.records.shape
        machine_times = problem.get_machine_times(x)

        fig = plt.figure()
        ax = fig.add_subplot(111)

        Y = np.flip(np.arange(n_machines))

        for i in range(n_machines):
            for j in range(n_jobs):
                width = problem.records[i][x[j]]
                left = machine_times[i][j]
                ax.barh(Y[i], width, left=left,
                        align='center', color='gray',
                        edgecolor='black', linewidth=0.8
                        )
                if label:
                    ax.text((left + width / 2), Y[i], str(x[j] + 1), ha='center', va='center', color='white',
                            fontsize=15)
        ax.set_xlabel("Time")
        ax.set_yticks(np.arange(n_machines))
        ax.set_yticklabels(["M%d" % (i + 1) for i in Y])
        ax.set_title("Makespan: %s" % np.round(problem.makespan(x), 3))
        if path is not None:
            plt.savefig(path)
        plt.show()



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.core.population import Population
from pymoo.operators.crossover.binx import mut_binomial
from pymoo.operators.crossover.expx import mut_exp
from pymoo.operators.repair.bounds_repair import is_out_of_bounds_by_problem, repair_random_init


def de_differential(X, F, dither=None, jitter=True, gamma=0.0001, return_differentials=False):
    n_parents, n_matings, n_var = X.shape
    assert n_parents % 2 == 1, "For the differential an odd number of values need to be provided"

    # make sure F is a one-dimensional vector
    F = np.ones(n_matings) * F

    # build the pairs for the differentials
    pairs = (np.arange(n_parents - 1) + 1).reshape(-1, 2)

    # the differentials from each pair subtraction
    diffs = np.zeros((n_matings, n_var))

    # for each difference
    for i, j in pairs:

        if dither == "vector":
            F = (F + np.random.random(n_matings) * (1 - F))
        elif dither == "scalar":
            F = F + np.random.random() * (1 - F)

        # http://www.cs.ndsu.nodak.edu/~siludwig/Publish/papers/SSCI20141.pdf
        if jitter:
            F = (F * (1 + gamma * (np.random.random(n_matings) - 0.5)))

        # an add the difference to the first vector
        diffs += F[:, None] * (X[i] - X[j])

    # now add the differentials to the first parent
    Xp = X[0] + diffs

    if return_differentials:
        return Xp, diffs
    else:
        return Xp


class DEX(Crossover):

    def __init__(self,
                 F=None,
                 CR=0.7,
                 variant="bin",
                 dither=None,
                 jitter=False,
                 n_diffs=1,
                 n_iter=1,
                 at_least_once=True,
                 **kwargs):

        super().__init__(1 + 2 * n_diffs, 1, **kwargs)
        self.n_diffs = n_diffs
        self.F = F
        self.CR = CR
        self.variant = variant
        self.at_least_once = at_least_once
        self.dither = dither
        self.jitter = jitter
        self.n_iter = n_iter

    def do(self, problem, pop, parents=None, **kwargs):

        # if a parents with array with mating indices is provided -> transform the input first
        if parents is not None:
            pop = [pop[mating] for mating in parents]

        # get the actual values from each of the parents
        X = np.swapaxes(np.array([[parent.get("X") for parent in mating] for mating in pop]), 0, 1).copy()

        n_parents, n_matings, n_var = X.shape

        # a mask over matings that need to be repeated
        m = np.arange(n_matings)

        # if the user provides directly an F value to use
        F = self.F if self.F is not None else rnd_F(m)

        # prepare the out to be set
        Xp = de_differential(X[:, m], F)

        # if the problem has boundaries to be considered
        if problem.has_bounds():

            for k in range(self.n_iter):
                # find the individuals which are still infeasible
                m = is_out_of_bounds_by_problem(problem, Xp)

                F = rnd_F(m)

                # actually execute the differential equation
                Xp[m] = de_differential(X[:, m], F)

            # if still infeasible do a random initialization
            Xp = repair_random_init(Xp, X[0], *problem.bounds())

        if self.variant == "bin":
            M = mut_binomial(n_matings, n_var, self.CR, at_least_once=self.at_least_once)
        elif self.variant == "exp":
            M = mut_exp(n_matings, n_var, self.CR, at_least_once=self.at_least_once)
        else:
            raise Exception(f"Unknown variant: {self.variant}")

        # take the first parents (this is already a copy)
        X = X[0]

        # set the corresponding values from the donor vector
        X[M] = Xp[M]

        return Population.new("X", X)


def rnd_F(m):
    return 0.5 * (1 + np.random.uniform(size=len(m)))



import warnings
from abc import abstractmethod

import numpy as np

# ---------------------------------------------------------------------------------------------------------
# Object Oriented Interface
# ---------------------------------------------------------------------------------------------------------

# ---- Abstract Class
from numpy.linalg import LinAlgError


class Normalization:

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def forward(self, X):
        pass

    @abstractmethod
    def backward(self, X):
        pass


# ---- Useful if normalization is optional - can be simply disabled by using this object
class NoNormalization(Normalization):

    def forward(self, X):
        return X

    def backward(self, X):
        return X


# ---- Normalizes between zero and one given bounds or estimating them
class ZeroToOneNormalization(Normalization):

    def __init__(self, xl=None, xu=None) -> None:
        super().__init__()

        # if both are None we are basically done because normalization is disabled
        if xl is None and xu is None:
            self.xl, self.xu = None, None
            return

        # if not set simply fall back no nan values
        if xl is None:
            xl = np.full_like(xu, np.nan)
        if xu is None:
            xu = np.full_like(xl, np.nan)

        xl, xu = np.copy(xl).astype(float), np.copy(xu).astype(float)

        # if both are equal then set the upper bound to none (always the 0 or lower bound will be returned then)
        xu[xl == xu] = np.nan

        # store the lower and upper bounds
        self.xl, self.xu = xl, xu

        # check out when the input values are nan
        xl_nan, xu_nan = np.isnan(xl), np.isnan(xu)

        # now create all the masks that are necessary
        self.xl_only, self.xu_only = np.logical_and(~xl_nan, xu_nan), np.logical_and(xl_nan, ~xu_nan)
        self.both_nan = np.logical_and(np.isnan(xl), np.isnan(xu))
        self.neither_nan = ~self.both_nan

        # if neither is nan than xu must be greater or equal than xl
        any_nan = np.logical_or(np.isnan(xl), np.isnan(xu))
        assert np.all(np.logical_or(xu >= xl, any_nan)), "xl must be less or equal than xu."

    def forward(self, X):
        if X is None or (self.xl is None and self.xu is None):
            return X

        xl, xu, xl_only, xu_only = self.xl, self.xu, self.xl_only, self.xu_only
        both_nan, neither_nan = self.both_nan, self.neither_nan

        # simple copy the input
        N = np.copy(X)

        # normalize between zero and one if neither of them is nan
        N[..., neither_nan] = (X[..., neither_nan] - xl[neither_nan]) / (xu[neither_nan] - xl[neither_nan])

        N[..., xl_only] = X[..., xl_only] - xl[xl_only]

        N[..., xu_only] = 1.0 - (xu[xu_only] - X[..., xu_only])

        return N

    def backward(self, N):
        if N is None or (self.xl is None and self.xu is None):
            return N

        xl, xu, xl_only, xu_only = self.xl, self.xu, self.xl_only, self.xu_only
        both_nan, neither_nan = self.both_nan, self.neither_nan

        X = N.copy()
        X[..., neither_nan] = xl[neither_nan] + N[..., neither_nan] * (xu[neither_nan] - xl[neither_nan])

        X[..., xl_only] = N[..., xl_only] + xl[xl_only]

        X[..., xu_only] = xu[xu_only] - (1.0 - N[..., xu_only])

        return X


# ---------------------------------------------------------------------------------------------------------
# Simple Normalization
# Does not consider any none values and assumes lower as well as upper bounds are given.
# ---------------------------------------------------------------------------------------------------------


class SimpleZeroToOneNormalization(Normalization):

    def __init__(self, xl=None, xu=None, estimate_bounds=True) -> None:
        super().__init__()
        self.xl = xl
        self.xu = xu
        self.estimate_bounds = estimate_bounds

    def forward(self, X):

        if self.estimate_bounds:
            if self.xl is None:
                self.xl = np.min(X, axis=0)
            if self.xu is None:
                self.xu = np.max(X, axis=0)

        xl, xu = self.xl, self.xu

        # if np.any(xl == xu):
        #     raise Exception("Normalization failed because lower and upper bounds are equal!")

        # calculate the denominator
        denom = xu - xl

        # we can not divide by zero -> plus small epsilon
        denom += (denom == 0) * 1e-32

        # normalize the actual values
        N = (X - xl) / denom

        return N

    def backward(self, X):
        return X * (self.xu - self.xl) + self.xl


# ---------------------------------------------------------------------------------------------------------
# Functional Interface
# ---------------------------------------------------------------------------------------------------------


def normalize(X, xl=None, xu=None, return_bounds=False, estimate_bounds_if_none=True):
    if estimate_bounds_if_none:
        if xl is None:
            xl = np.min(X, axis=0)
        if xu is None:
            xu = np.max(X, axis=0)

    if isinstance(xl, float) or isinstance(xl, int):
        xl = np.full(X.shape[-1], xl)

    if isinstance(xu, float) or isinstance(xu, int):
        xu = np.full(X.shape[-1], xu)

    norm = ZeroToOneNormalization(xl, xu)
    X = norm.forward(X)

    if not return_bounds:
        return X
    else:
        return X, norm.xl, norm.xu


def denormalize(x, xl, xu):
    return ZeroToOneNormalization(xl, xu).backward(x)


def standardize(x, return_bounds=False):
    mean = np.mean(x, axis=0)
    std = np.std(x, axis=0)

    # standardize
    val = (x - mean) / std

    if not return_bounds:
        return val
    else:
        return val, mean, std


def destandardize(x, mean, std):
    return (x * std) + mean


# ---------------------------------------------------------------------------------------------------------
# Pre Normalization
# A class inheriting from it can use the in-built feature of normalizing
# ---------------------------------------------------------------------------------------------------------


class PreNormalization:

    def __init__(self, zero_to_one=False, ideal=None, nadir=None, **kwargs):

        # normalization related stuff if that should be performed beforehand
        self.ideal, self.nadir = ideal, nadir

        if zero_to_one:
            assert self.ideal is not None and self.nadir is not None, "For normalization either provide pf or bounds!"

            n_dim = len(self.ideal)
            self.normalization = ZeroToOneNormalization(self.ideal, self.nadir)

            # now the ideal and nadir points have change to only zeros and ones
            self.ideal, self.nadir = np.zeros(n_dim), np.ones(n_dim)

        else:
            self.normalization = NoNormalization()

    def do(self, *args, **kwargs):
        pass


# ---------------------------------------------------------------------------------------------------------
# Normalization in the Objective Space
# ---------------------------------------------------------------------------------------------------------


def find_ideal(F, current=None):
    p = F.min(axis=0)
    if current is not None:
        p = np.minimum(current, p)
    return p


def get_extreme_points_c(F, ideal_point, extreme_points=None):
    # calculate the asf which is used for the extreme point decomposition
    weights = np.eye(F.shape[1])
    weights[weights == 0] = 1e6

    # add the old extreme points to never loose them for normalization
    _F = F
    if extreme_points is not None:
        _F = np.concatenate([extreme_points, _F], axis=0)

    # use __F because we substitute small values to be 0
    __F = _F - ideal_point
    __F[__F < 1e-3] = 0

    # update the extreme points for the normalization having the highest asf value each
    F_asf = np.max(__F * weights[:, None, :], axis=2)

    I = np.argmin(F_asf, axis=1)
    extreme_points = _F[I, :]

    return extreme_points


def get_nadir_point(extreme_points, ideal_point, worst_point, worst_of_front, worst_of_population):
    try:

        # find the intercepts using gaussian elimination
        M = extreme_points - ideal_point
        b = np.ones(extreme_points.shape[1])
        plane = np.linalg.solve(M, b)

        warnings.simplefilter("ignore")
        intercepts = 1 / plane

        nadir_point = ideal_point + intercepts

        # check if the hyperplane makes sense
        if not np.allclose(np.dot(M, plane), b) or np.any(intercepts <= 1e-6):
            raise LinAlgError()

        # if the nadir point should be larger than any value discovered so far set it to that value
        # NOTE: different to the proposed version in the paper
        b = nadir_point > worst_point
        nadir_point[b] = worst_point[b]

    except LinAlgError:

        # fall back to worst of front otherwise
        nadir_point = worst_of_front

    # if the range is too small set it to worst of population
    b = nadir_point - ideal_point <= 1e-6
    nadir_point[b] = worst_of_population[b]

    return nadir_point


class ObjectiveSpaceNormalization:

    def __init__(self) -> None:
        super().__init__()
        self._ideal = None
        self._infeas_ideal = None
        self._worst = None

    def update(self, pop):
        F, feas = pop.get("F", "feasible")
        self._infeas_ideal = find_ideal(F, current=self._infeas_ideal)

        if np.any(feas):
            self._ideal = find_ideal(F[feas[:, 0]], self._ideal)

    def ideal(self, only_feas=True):
        return self._ideal if only_feas else self._infeas_ideal



# The code is translated from Matlab by ahcheriet@gmail.com
# ========================================================|#
# The 14 test functions are for cec2018 competition on    |
# dynamic multiobjective optimisation. This document is   |
# free to disseminate for academic use.                   |
# --------------------------------------------------------|#
# The "time" term in the test suite is defined as:        |
#          t=1/nt*floor(tau/taut)                         |
# where - nt:    severity of change                       |
#       - taut:  frequency of change                      |
#       - tau:   current generation counter               |
# --------------------------------------------------------|#
# Any questions can be directed to                        |
#    Dr. Shouyong Jiang at math4neu@gmail.com.            |
#                                                         |
# ========================================================|#

# cec2018_DF(probID, x, tau, taut, nt)
# INPUT:
#       probID: test problem identifier (i.e. 'DF1')
#       x:      variable vector
#       tau:    current generation counter
#       taut:   frequency of change
#       nt:     severity of change
#
# OUTPUT:
#       f:      objective vector
#
from numpy import power, setdiff1d, exp, prod


def get_bounds(problem_id='DF1', n_vars=10):
    lx = np.zeros(n_vars)
    ux = np.ones(n_vars)
    if problem_id == 'DF3':
        lx = -1 * np.ones(n_vars)
        ux = 2 * np.ones(n_vars)
        lx[0] = 0.0
        ux[0] = 1.0
    if problem_id == 'DF4':
        lx = -2 * np.ones(n_vars)
        ux = -2 * np.ones(n_vars)
    if problem_id == 'DF5':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        ux[0] = 1.0
    if problem_id == 'DF6':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        ux[0] = 1.0
    if problem_id == 'DF7':
        lx = 0 * np.ones(n_vars)
        ux = 1 * np.ones(n_vars)
        lx[0] = 1.0
        ux[0] = 4.0
    if problem_id == 'DF8':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        ux[0] = 1.0
    if problem_id == 'DF9':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        ux[0] = 1.0
    if problem_id == 'DF10':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        lx[1] = 0.0
        ux[0] = 1.0
        ux[1] = 1.0
    if problem_id == 'DF12':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        lx[1] = 0.0
        ux[0] = 1.0
        ux[1] = 1.0
    if problem_id == 'DF13':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        lx[1] = 0.0
        ux[0] = 1.0
        ux[1] = 1.0
    if problem_id == 'DF14':
        lx = -1 * np.ones(n_vars)
        ux = -1 * np.ones(n_vars)
        lx[0] = 0.0
        lx[1] = 0.0
        ux[0] = 1.0
        ux[1] = 1.0
    return lx, ux


def cec2018_DF(problemID='DF1', x=None, t=None):
    # INPUT:
    #       probID: test problem identifier (i.e. 'DF1')
    #       x:      variable vector
    #       tau:    current generation counter
    #       taut:   frequency of change
    #       nt:     severity of change

    # OUTPUT:
    #       f:      objective vector

    # the first change occurs after T0 generations, that is, the
    # generation at which a change occurs is (T0+1), (T0+taut+1), etc.

    T0 = 50
    # calculate time instant
    n = len(x)
    f = {}
    if problemID == 'DF1':
        G = abs(sin(0.5 * pi * t))
        H = 0.75 * sin(0.5 * pi * t) + 1.25
        g = 1 + sum((x[1:] - G) ** 2)
        f[0] = x[0]
        f[1] = g * power(1 - (x[0] / g), H)
    if problemID == 'DF2':
        G = abs(sin(0.5 * pi * t))
        r = 1 + floor((n - 1) * G)
        tmp = setdiff1d(range(0, n), [int(r)])
        g = 1 + sum([(x[int(index)] - G) ** 2 for index in tmp])
        f[0] = x[int(r - 1)]
        f[1] = g * (power(1 - (x[int(r - 1)] / g), 0.5))
    if problemID == 'DF3':
        G = sin(0.5 * pi * t)
        H = G + 1.5
        g = 1 + sum(power(x[1:] - G - x[0], H) ** 2)
        f[0] = x[0]
        f[1] = g * power(1 - (x[0] / g), H)
    if problemID == 'DF4':
        a = sin(0.5 * pi * t)
        b = 1 + abs(cos(0.5 * pi * t))
        H = 1.5 + a
        g = 1 + sum((x[1:] - a * x[0] ** 2 / x[1:]) ** 2)
        f[0] = g * power(abs(x[0] - a), H)
        f[1] = g * power(abs(x[0] - a - b), H)
    if problemID == 'DF5':
        G = sin(0.5 * pi * t)
        w = floor(10 * G)
        g = 1 + sum((x[1:] - G) ** 2)
        f[0] = g * (x[0] + 0.02 * sin(w * pi * x[0]))
        f[1] = g * (1 - x[0] + 0.02 * sin(w * pi * x[0]))
    if problemID == 'DF6':
        G = sin(0.5 * pi * t)
        a = 0.2 + 2.8 * abs(G)
        y = x[1:] - G
        g = 1 + sum((abs(G) * y ** 2 - 10 * cos(2 * pi * y) + 10))
        f[0] = g * power(x[0] + 0.1 * sin(3 * pi * x[0]), a)
        f[1] = g * power(1 - x[0] + 0.1 * sin(3 * pi * x[0]), a)
    if problemID == 'DF7':
        a = 5 * cos(0.5 * pi * t)
        tmp = 1 / (1 + exp(a * (x[0] - 2.5)))
        g = 1 + sum(power(x[1:] - tmp, 2))
        f[0] = g * (1 + t) / x[0]
        f[1] = g * x[0] / (1 + t)
    if problemID == 'DF8':
        G = sin(0.5 * pi * t)
        a = 2.25 + 2 * cos(2 * pi * t)
        b = 100 * G ** 2
        tmp = G * sin(power(4 * pi * x[0], b)) / (1 + abs(G))
        g = 1 + sum((x[1:] - tmp) ** 2)
        f[0] = g * (x[0] + 0.1 * sin(3 * pi * x[0]))
        f[1] = g * power(1 - x[1] + 0.1 * sin(3 * pi * x[1]), a)
    if problemID == 'DF9':
        N = 1 + floor(10 * abs(sin(0.5 * pi * t)))
        g = 1
        for i in range(1, n):
            tmp = x[i] - cos(4 * t + x[0] + x[i - 1])
            g = g + tmp ** 2
        f[0] = g * (x[0] + max(0, (0.1 + 0.5 / N) * sin(2 * N * pi * x[0])))
        f[1] = g * (1 - x[0] + max(0, (0.1 + 0.5 / N) * sin(2 * N * pi * x[0])))
    if problemID == 'DF10':
        G = sin(0.5 * pi * t)
        H = 2.25 + 2 * cos(0.5 * pi * t)
        tmp = sin(2 * pi * (x[0] + x[1])) / (1 + abs(G))
        g = 1 + sum((x[2:] - tmp) ** 2)
        f[0] = g * power(sin(0.5 * pi * x[0]), H)
        f[1] = g * power(sin(0.5 * pi * x[1]), H) * power(cos(0.5 * pi * x[0]), H)
        f[2] = g * power(cos(0.5 * pi * x[1]), H) * power(cos(0.5 * pi * x[0]), H)
    if problemID == 'DF11':
        G = abs(sin(0.5 * pi * t))
        g = 1 + G + sum((x[2:] - 0.5 * G * x[0]) ** 2)
        y = [pi * G / 6.0 + (pi / 2 - pi * G / 3.0) * x[i] for i in [0, 1]]
        f[0] = g * sin(y[0])
        f[1] = g * sin(y[1]) * cos(y[0])
        f[2] = g * cos(y[1]) * cos(y[0])
    if problemID == 'DF12':
        k = 10 * sin(pi * t)
        tmp1 = x[2:] - sin(t * x[0])
        tmp2 = [sin(floor(k * (2 * x[0] - 1)) * pi / 2)]
        g = 1 + sum(tmp1 ** 2) + prod(tmp2)
        f[0] = g * cos(0.5 * pi * x[1]) * cos(0.5 * pi * x[0])
        f[1] = g * sin(0.5 * pi * x[1]) * cos(0.5 * pi * x[0])
        f[2] = g * sin(0.5 * pi * x[1])
    if problemID == 'DF13':
        G = sin(0.5 * pi * t);
        p = floor(6 * G);
        g = 1 + sum((x[2:] - G) ** 2)
        f[0] = g * cos(0.5 * pi * x[0]) ** 2
        f[1] = g * cos(0.5 * pi * x[1]) ** 2
        f[2] = g * sin(0.5 * pi * x[0]) ** 2 + sin(0.5 * pi * x[0]) * cos(p * pi * x[0]) ** 2 + sin(
            0.5 * pi * x[1]) ** 2 + sin(0.5 * pi * x[1]) * cos(p * pi * x[1]) ** 2
    if problemID == 'DF14':
        G = sin(0.5 * pi * t)
        g = 1 + sum((x[2:] - G) ** 2)
        y = 0.5 + G * (x[0] - 0.5)
        f[0] = g * (1 - y + 0.05 * sin(6 * pi * y))
        f[1] = g * (1 - x[1] + 0.05 * sin(6 * pi * x[1])) * (y + 0.05 * sin(6 * pi * y))
        f[2] = g * (x[1] + 0.05 * sin(6 * pi * x[1])) * (y + 0.05 * sin(6 * pi * y))
    return f


import numpy as np
from numpy import pi, dot, floor, sin, cos, multiply, arange
from copy import copy

false = False
true = True


# cec2018_pf.m

# ========================================================|#
# PF calculation for 14 cec2018 test functions on         |
# dynamic multiobjective optimisation. This document is   |
# free to disseminate for academic use.                   |
# --------------------------------------------------------|#
# The "time" term in the test suite is defined as:        |
#          t=1/nt*floor(tau/taut)                         |
# where - nt:    severity of change                       |
#       - taut:  frequency of change                      |
#       - tau:   current generation counter               |
# --------------------------------------------------------|#
# Any questions can be directed to                        |
#    Dr. Shouyong Jiang at math4neu@gmail.com.            |
#                                                         |
# ========================================================|#


def cec2018_DF_PF(probID=None, t=1, n_points=100, *args, **kwargs):
    # INPUT:
    #       probID: test problem identifier (i.e. 'DF1')
    #       tau:    current generation counter
    #       taut:   frequency of change
    #       nt:     severity of change

    # OUTPUT:
    #       h:      nondominated solutions

    T0 = 50
    g = 1
    H = 50

    if 'DF1' == (probID):
        x = np.linspace(0, 1, n_points)
        H = dot(0.75, sin(dot(dot(0.5, pi), t))) + 1.25
        f1 = copy(x)
        f2 = dot(g, (1 - (x / g) ** H))
        h = get_PF(np.array([f1, f2]), false)
    if 'DF2' == (probID):
        x = np.linspace(0, 1, n_points)
        G = abs(sin(dot(dot(0.5, pi), t)))
        f1 = copy(x)
        f2 = dot(g, (1 - (x / g) ** 0.5))  # To be sure
        h = get_PF(np.array([f1, f2]), false)
    if 'DF3' == (probID):
        x = np.linspace(0, 1, n_points)
        G = sin(dot(dot(0.5, pi), t))
        H = G + 1.5
        f1 = copy(x)
        f2 = dot(g, (1 - (x / g) ** H))
        h = get_PF(np.array([f1, f2]), false)
    if 'DF4' == (probID):
        a = sin(dot(dot(0.5, pi), t))
        b = 1 + abs(cos(dot(dot(0.5, pi), t)))
        x = np.linspace(a, a + b)
        H = 1.5 + a
        f1 = dot(g, abs(x - a) ** H)
        f2 = dot(g, abs(x - a - b) ** H)  # Maybe
        h = get_PF(np.array([f1, f2]), false)
    if 'DF5' == (probID):
        x = np.linspace(0, 1, n_points)
        G = sin(dot(dot(0.5, pi), t))
        w = floor(dot(10, G))
        f1 = dot(g, (x + dot(0.02, sin(dot(dot(w, pi), x)))))
        f2 = dot(g, (1 - x + dot(0.02, sin(dot(dot(w, pi), x)))))
        h = get_PF(np.array([f1, f2]), false)
    if 'DF6' == (probID):
        x = np.linspace(0, 1, n_points)
        G = sin(dot(dot(0.5, pi), t))
        a = 0.2 + dot(2.8, abs(G))
        f1 = dot(g, (x + dot(0.1, sin(dot(dot(3, pi), x)))) ** a)
        f2 = dot(g, (1 - x + dot(0.1, sin(dot(dot(3, pi), x)))) ** a)
        h = get_PF(np.array([f1, f2]), false)
    if 'DF7' == (probID):
        x = np.linspace(1, 4, n_points)
        f1 = dot(g, (1 + t)) / x
        f2 = dot(g, x) / (1 + t)
        h = get_PF(np.array([f1, f2]), false)
    if 'DF8' == (probID):
        x = np.linspace(0, 1, n_points)
        a = 2.25 + dot(2, cos(dot(dot(2, pi), t)))
        f1 = dot(g, (x + dot(0.1, sin(dot(dot(3, pi), x)))))
        f2 = dot(g, (1 - x + dot(0.1, sin(dot(dot(3, pi), x)))) ** a)
        h = get_PF(np.array([f1, f2]), false)
    if 'DF9' == (probID):
        x = np.linspace(0, 1, n_points)
        #        N = 1 + floor(dot(10, abs(sin(dot(dot(0.5, pi), t)))))

        N = 1 + floor(10 * abs(sin(0.5 * pi * t)))
        print(max(0, max((0.1 + 0.5 / N) * sin(2 * N * pi * x))))
        f1 = g * (x + max(0, max((0.1 + 0.5 / N) * sin(2 * N * pi * x))))
        f2 = g * (1 - x + max(0, max((0.1 + 0.5 / N) * sin(2 * N * pi * x))))

        #        f1 = dot(g, (x + max(0, dot((0.1 + 0.5 / N), sin(dot(dot(dot(2, N), pi), x))))))
        #        f2 = dot(g, (1 - x + max(0, dot((0.1 + 0.5 / N), sin(dot(dot(dot(2, N), pi), x))))))
        h = get_PF(np.array([f1, f2]), true)
    if 'DF10' == (probID):
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        H = 2.25 + dot(2, cos(dot(dot(0.5, pi), t)))
        f1 = dot(g, sin(dot(dot(0.5, pi), x1)) ** H)
        f2 = multiply(dot(g, sin(dot(dot(0.5, pi), x2)) ** H),
                      cos(dot(dot(0.5, pi), x1)) ** H)
        f3 = multiply(dot(g, cos(dot(dot(0.5, pi), x2)) ** H),
                      cos(dot(dot(0.5, pi), x1)) ** H)
        h = get_PF(np.array([f1, f2, f3]), false)
    if 'DF11' == (probID):
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        G = abs(sin(dot(dot(0.5, pi), t)))
        y1 = dot(pi, G) / 6 + dot((pi / 2 - dot(pi, G) / 3), x1)
        y2 = dot(pi, G) / 6 + dot((pi / 2 - dot(pi, G) / 3), x2)
        f1 = multiply(g, sin(y1))
        f2 = dot(multiply(g, sin(y2)), cos(y1))
        f3 = dot(multiply(g, cos(y2)), cos(y1))
        h = get_PF(np.array([f1, f2, f3]), false)
    if 'DF12' == (probID):
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        k = dot(10, sin(dot(pi, t)))
        tmp2 = abs(
            multiply(sin(dot(floor(dot(k, (dot(2, x1) - 1))), pi) / 2),
                     sin(dot(floor(dot(k, (dot(2, x2) - 1))), pi) / 2)))
        g = 1 + tmp2
        f1 = multiply(multiply(g, cos(dot(dot(0.5, pi), x2))),
                      cos(dot(dot(0.5, pi), x1)))
        f2 = multiply(multiply(g, sin(dot(dot(0.5, pi), x2))),
                      cos(dot(dot(0.5, pi), x1)))
        f3 = multiply(g, sin(dot(dot(0.5, pi), x1)))
        h = get_PF(np.array([f1, f2, f3]), true)
    if 'DF13' == (probID):
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        G = sin(dot(dot(0.5, pi), t))
        p = floor(dot(6, G))
        f1 = multiply(g, cos(dot(dot(0.5, pi), x1)) ** 2)
        f2 = multiply(g, cos(dot(dot(0.5, pi), x2)) ** 2)
        f3 = multiply(g, sin(dot(dot(0.5, pi), x1)) ** 2) + multiply(
            sin(dot(dot(0.5, pi), x1)),
            cos(dot(dot(p, pi), x1)) ** 2) + sin(
            dot(dot(0.5, pi), x2)) ** 2 + multiply(
            sin(dot(dot(0.5, pi), x2)), cos(dot(dot(p, pi), x2)) ** 2)
        h = get_PF(np.array([f1, f2, f3]), true)
    if 'DF14' == (probID):
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        G = sin(dot(dot(0.5, pi), t))
        y = 0.5 + dot(G, (x1 - 0.5))
        f1 = multiply(g, (1 - y + dot(0.05, sin(dot(dot(6, pi), y)))))
        f2 = multiply(multiply(g, (
                1 - x2 + dot(0.05, sin(dot(dot(6, pi), x2))))),
                      (y + dot(0.05, sin(dot(dot(6, pi), y)))))
        f3 = multiply(
            multiply(g, (x2 + dot(0.05, sin(dot(dot(6, pi), x2))))),
            (y + dot(0.05, sin(dot(dot(6, pi), y)))))
        h = get_PF(np.array([f1, f2, f3]), false)
    return h


def get_PF(f=None, nondominate=None, *args, **kwargs):
    ncell = len(f)
    s = np.size(f[1])
    h = []
    for i in arange(ncell):
        fi = np.reshape(f[i], s, order='F')
        h.append(fi)
    h = np.array(h).T
    h = np.reshape(h, (s, ncell))

    if nondominate:
        print("Run Non dominating Sorting")
        h = []
        pass
    #     in_ = get_skyline(h)
    #     h = h(in_, arange())
    return h



import abc

import numpy as np

from pymoo.core.population import Population
from pymoo.core.repair import Repair


def is_in_bounds(X, xl, xu):
    return np.where(np.all(np.logical_and(X >= xl, X <= xu), axis=1))[0]


def is_in_bounds_by_problem(problem, X):
    return is_in_bounds(X, problem.xl, problem.xu)


def is_out_of_bounds(X, xl, xu):
    return np.where(np.any(np.logical_or(X < xl, X > xu), axis=1))[0]


def is_out_of_bounds_by_problem(problem, X):
    return is_out_of_bounds(X, problem.xl, problem.xu)


def repeat_bounds(xl, xu, n):
    XL = np.tile(xl, (n, 1))
    XU = np.tile(xu, (n, 1))
    return XL, XU


def repair_clamp(Xp, xl, xu):
    XL, XU = repeat_bounds(xl, xu, len(Xp))

    I = np.where(Xp < XL)
    Xp[I] = XL[I]

    I = np.where(Xp > XU)
    Xp[I] = XU[I]

    return Xp


def repair_periodic(Xp, xl, xu):
    XL, XU = repeat_bounds(xl, xu, len(Xp))

    S = (XU - XL)

    I = np.where(Xp < XL)
    Xp[I] = XU[I] - (XL[I] - Xp[I]) % S[I]

    I = np.where(Xp > XU)
    Xp[I] = XL[I] + (Xp[I] - XU[I]) % S[I]

    return Xp


def repair_random_init(Xp, X, xl, xu):
    XL, XU = repeat_bounds(xl, xu, len(Xp))

    i, j = np.where(Xp < XL)
    if len(i) > 0:
        Xp[i, j] = XL[i, j] + np.random.random(len(i)) * (X[i, j] - XL[i, j])

    i, j = np.where(Xp > XU)
    if len(i) > 0:
        Xp[i, j] = XU[i, j] - np.random.random(len(i)) * (XU[i, j] - X[i, j])

    return Xp


class BoundsRepair(Repair):

    def _do(self,
            problem,
            pop_or_X,
            check_out_of_bounds=True,
            **kwargs):

        is_array = not isinstance(pop_or_X, Population)

        X = pop_or_X if is_array else pop_or_X.get("X")

        X = self.repair_out_of_bounds(problem, X, **kwargs)

        assert len(is_out_of_bounds_by_problem(problem, X)) == 0

        if is_array:
            return X
        else:
            pop_or_X.set("X", X)
            return pop_or_X

    @abc.abstractmethod
    def repair_out_of_bounds(self, problem, X, **kwargs):
        pass



from pymoo.core.decision_making import DecisionMaking
from pymoo.util.normalization import normalize


class CompromiseProgramming(DecisionMaking):

    def __init__(self, metric="euclidean", **kwargs) -> None:
        super().__init__(**kwargs)
        self.metric = metric

    def _do(self, F, **kwargs):

        F, _, ideal_point, nadir_point = normalize(F,
                                                   xl=self.ideal_point,
                                                   xu=self.nadir_point,
                                                   estimate_bounds_if_none=True,
                                                   return_bounds=True)

        return None



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.problems.many.dtlz import DTLZ1, DTLZ2, DTLZ3, DTLZ4, get_ref_dirs


class C1DTLZ1(DTLZ1):

    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, n_ieq_constr=1, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        super()._evaluate(X, out, *args, **kwargs)
        out["G"] = constraint_c1_linear(out["F"])

    def _calc_pareto_front(self, ref_dirs=None, *args, **kwargs):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        return super()._calc_pareto_front(ref_dirs, *args, **kwargs)


class C1DTLZ3(DTLZ3):

    def __init__(self, n_var=12, n_obj=3, r=None, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, n_ieq_constr=1, **kwargs)

        if r is None:
            if self.n_obj < 5:
                r = 9.0
            elif 5 <= self.n_obj <= 12:
                r = 12.5
            else:
                r = 15.0

        self.r = r

    def _evaluate(self, X, out, *args, **kwargs):
        super()._evaluate(X, out, *args, **kwargs)
        out["G"] = constraint_c1_spherical(out["F"], self.r)

    def _calc_pareto_front(self, ref_dirs=None, *args, **kwargs):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        return super()._calc_pareto_front(ref_dirs, *args, **kwargs)


class C2DTLZ2(DTLZ2):

    def __init__(self, n_var=12, n_obj=3, r=None, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, n_ieq_constr=1, **kwargs)

        if r is None:
            if n_obj == 2:
                r = 0.2
            elif n_obj == 3:
                r = 0.4
            else:
                r = 0.5

        self.r = r

    def _evaluate(self, X, out, *args, **kwargs):
        super()._evaluate(X, out, *args, **kwargs)
        out["G"] = constraint_c2(out["F"], self.r)

    def _calc_pareto_front(self, ref_dirs=None, *args, **kwargs):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        F = super()._calc_pareto_front(ref_dirs, *args, **kwargs)
        G = constraint_c2(F, r=self.r)
        G[G <= 0] = 0
        return F[G <= 0]


class C3DTLZ1(DTLZ1):

    def __init__(self, n_var=12, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, n_ieq_constr=n_obj, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        super()._evaluate(X, out, *args, **kwargs)
        out["G"] = constraint_c3_linear(out["F"])


class C3DTLZ4(DTLZ4):

    def __init__(self, n_var=7, n_obj=3, **kwargs):
        super().__init__(n_var=n_var, n_obj=n_obj, n_ieq_constr=n_obj, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        super()._evaluate(X, out, *args, **kwargs)
        out["G"] = constraint_c3_spherical(out["F"])

    def _calc_pareto_front(self, ref_dirs=None, *args, **kwargs):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        F = super()._calc_pareto_front(ref_dirs, *args, **kwargs)
        a = np.sqrt(np.sum(F ** 2, 1) - 3 / 4 * np.max(F ** 2, axis=1))
        a = np.expand_dims(a, axis=1)
        a = np.tile(a, [1, ref_dirs.shape[1]])
        F = F / a

        return F


def constraint_c1_linear(f):
    g = - (1 - f[:, -1] / 0.6 - anp.sum(f[:, :-1] / 0.5, axis=1))
    return g


def constraint_c1_spherical(f, r):
    radius = anp.sum(f ** 2, axis=1)
    g = - (radius - 16) * (radius - r ** 2)

    return g


def constraint_c2(f, r):
    n_obj = f.shape[1]

    v1 = anp.full(f.shape[0], anp.inf)

    for i in range(n_obj):
        temp = (f[:, i] - 1) ** 2 + (anp.sum(f ** 2, axis=1) - f[:, i] ** 2) - r ** 2
        v1 = anp.minimum(temp.flatten(), v1)

    a = 1 / anp.sqrt(n_obj)
    v2 = anp.sum((f - a) ** 2, axis=1) - r ** 2
    g = anp.minimum(v1, v2.flatten())

    return g


def constraint_c3_linear(f):  # M lines
    n_obj = f.shape[1]
    g = []

    for i in range(n_obj):
        _g = 1 - f[:, i] / 0.5 - (anp.sum(f, axis=1) - f[:, i])
        g.append(_g)

    return anp.column_stack(g)


def constraint_c3_spherical(f):  # M ellipse
    n_obj = f.shape[1]
    g = []

    for i in range(n_obj):
        _g = 1 - f[:, i] ** 2 / 4 - (anp.sum(f ** 2, axis=1) - f[:, i] ** 2)
        g.append(_g)
    return anp.column_stack(g)


def constraint_c4_cylindrical(f, r):  # cylindrical
    l = anp.mean(f, axis=1)
    l = anp.expand_dims(l, axis=1)
    g = -anp.sum(anp.power(f - l, 2), axis=1) + anp.power(r, 2)
    return g


import pymoo.gradient.toolbox as anp

from pymoo.core.problem import Problem


class SimpleMultiModal01(Problem):

    def __init__(self):
        super().__init__(n_var=2, n_obj=1, n_ieq_constr=0, xl=0, xu=1, vtype=float)

    def _evaluate(self, X, out, *args, **kwargs):
        x, y = X[:, 0], X[:, 1]
        out["F"] = - anp.exp(-9 * abs(x * y)) * anp.sin(3 * anp.pi * x) ** 2 * anp.sin(3 * anp.pi * y) ** 2



# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True


import numpy as np


def stochastic_ranking(double[:] f, double[:] phi, double pr, long[:] I=None):
    if I is None:
        I = np.arange(len(f))
    return np.array(c_stochastic_ranking(f, phi, pr, I))



def c_stochastic_ranking(double[:] f, double[:] phi, double pr, long[:] I):
    cdef int _lambda, i, j, at_least_one_swap
    cdef double u

    _lambda = len(f)

    for i in range(_lambda):

        at_least_one_swap = 0

        for j in range(_lambda - 1):

            u = np.random.random()

            if u < pr or (phi[I[j]] == 0 and phi[I[j + 1]] == 0):
                if f[I[j]] > f[I[j + 1]]:
                    c_swap(I, j, j + 1)
                    at_least_one_swap = 1
            else:
                if phi[I[j]] > phi[I[j + 1]]:
                    c_swap(I, j, j + 1)
                    at_least_one_swap = 1

        if at_least_one_swap == 0:
            break

    return I


cdef c_swap(long[:] x, int a, int b):
    cdef int tmp
    tmp = x[a]
    x[a] = x[b]
    x[b] = tmp




import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.distance import cdist

from pymoo.core.problem import ElementwiseProblem


class TravelingSalesman(ElementwiseProblem):

    def __init__(self, cities, **kwargs):
        """
        A two-dimensional traveling salesman problem (TSP)

        Parameters
        ----------
        cities : numpy.array
            The cities with 2-dimensional coordinates provided by a matrix where where city is represented by a row.

        """
        n_cities, _ = cities.shape

        self.cities = cities
        self.D = cdist(cities, cities)

        super(TravelingSalesman, self).__init__(
            n_var=n_cities,
            n_obj=1,
            xl=0,
            xu=n_cities,
            vtype=int,
            **kwargs
        )

    def _evaluate(self, x, out, *args, **kwargs):
        out['F'] = self.get_route_length(x)

    def get_route_length(self, x):
        n_cities = len(x)
        dist = 0
        for k in range(n_cities - 1):
            i, j = x[k], x[k + 1]
            dist += self.D[i, j]

        last, first = x[-1], x[0]
        dist += self.D[last, first]  # back to the initial city
        return dist


def create_random_tsp_problem(n_cities, grid_width=100.0, grid_height=None, seed=None):
    if seed is not None:
        np.random.seed(seed)
    grid_height = grid_height if grid_height is not None else grid_width
    cities = np.random.random((n_cities, 2)) * [grid_width, grid_height]
    return TravelingSalesman(cities)


def visualize(problem, x, fig=None, ax=None, show=True, label=True):
    with plt.style.context('ggplot'):

        if fig is None or ax is None:
            fig, ax = plt.subplots()

        # plot cities using scatter plot
        ax.scatter(problem.cities[:, 0], problem.cities[:, 1], s=250)
        if label:
            # annotate cities
            for i, c in enumerate(problem.cities):
                ax.annotate(str(i), xy=c, fontsize=10, ha="center", va="center", color="white")

        # plot the line on the path
        for i in range(len(x)):
            current = x[i]
            next_ = x[(i + 1) % len(x)]
            ax.plot(problem.cities[[current, next_], 0], problem.cities[[current, next_], 1], 'r--')

        fig.suptitle("Route length: %.4f" % problem.get_route_length(x))

        if show:
            fig.show()



import math

import numpy as np

from pymoo.core.selection import Selection
from pymoo.util.misc import random_permuations


class RandomSelection(Selection):

    def _do(self, _, pop, n_select, n_parents, **kwargs):
        # number of random individuals needed
        n_random = n_select * n_parents

        # number of permutations needed
        n_perms = math.ceil(n_random / len(pop))

        # get random permutations and reshape them
        P = random_permuations(n_perms, len(pop))[:n_random]

        return np.reshape(P, (n_select, n_parents))


def fast_fill_random(X, N, columns=None, Xp=None, n_max_attempts=10):
    """

    Parameters
    ----------
    X : np.ndarray
        The actually array to fill with random values.
    N : int
        The upper limit for the values. The values will be in range (0, ..., N)
    columns : list
        The columns which should be filled randomly. Other columns indicate duplicates
    Xp : np.ndarray
        If some other duplicates shall be avoided by default

    """

    _, n_cols = X.shape

    if columns is None:
        columns = range(n_cols)

    # all columns set so far to be checked for duplicates
    J = []

    # for each of the columns which should be set to be no duplicates
    for col in columns:

        D = X[:, J]
        if Xp is not None:
            D = np.column_stack([D, Xp])

        # all the remaining indices that need to be filled with no duplicates
        rem = np.arange(len(X))

        for _ in range(n_max_attempts):

            if len(rem) > N:
                X[rem, col] = np.random.choice(N, replace=True, size=len(rem))
            else:
                X[rem, col] = np.random.permutation(N)[:len(rem)]

            rem = np.where((X[rem, col][:, None] == D[rem]).any(axis=1))[0]

            if len(rem) == 0:
                break

        J.append(col)

    return X



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class Rosenbrock(Problem):
    def __init__(self, n_var=2):
        super().__init__(n_var=n_var, n_obj=1, n_ieq_constr=0, xl=-2.048, xu=2.048, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        l = []
        for i in range(x.shape[1] - 1):
            val = 100 * (x[:, i + 1] - x[:, i] ** 2) ** 2 + (1 - x[:, i]) ** 2
            l.append(val)
        out["F"] = anp.sum(anp.column_stack(l), axis=1)

    def _calc_pareto_front(self):
        return 0.0

    def _calc_pareto_set(self):
        return np.full(self.n_var, 1.0)



import warnings

import autograd.numpy as anp
import numpy as np
from autograd.core import VJPNode, vspace, backward_pass
from autograd.tracer import new_box, isbox


def run_and_trace(f, x):
    start_node = VJPNode.new_root()

    start_box = new_box(x, 0, start_node)
    out = f(start_box)

    return out, start_box


def autograd_elementwise_value_and_grad(f, x):
    out, pullback = run_and_trace(f, x)

    jac = dict()
    for name in out:
        val = out[name]

        if val is not None:

            if len(val.shape) == 0:
                val = anp.array([val])

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                # the backward pass is done for each objective function once
                grad = []
                for j in range(len(val)):
                    b = np.zeros(val.shape)
                    b[j] = 1

                    n = new_box(b, 0, VJPNode.new_root())
                    _grad = backward_pass(n, val._node)
                    grad.append(_grad)

                out[name] = np.array(val._value)
                jac[name] = anp.stack(grad, axis=0)._value

    return out, jac


def autograd_vectorized_value_and_grad(f, x):
    end, start = run_and_trace(f, x)

    out, jac = dict(), dict()

    end = {k: v for k, v in end.items() if v is not None}

    for name, val in end.items():

        v = val
        if hasattr(v, "_value"):
            v = np.array(v._value)
        out[name] = v

        # if the end_box is not a box - autograd can not track back
        if not isbox(val):
            n, m = val.shape
            jac[name] = np.zeros((n, m, x.shape[1]))

        else:

            # the backward pass is done for each objective function once
            grad = []
            for j in range(val.shape[1]):
                b = anp.zeros(val.shape)
                b[:, j] = 1
                n = new_box(b, 0, VJPNode.new_root())
                _grad = backward_pass(n, val._node)
                grad.append(_grad)

            jac[name] = anp.stack(grad, axis=1)._value

    return out, jac



import numpy as np

from pymoo.util.dominator import Dominator
from pymoo.util.function_loader import load_function


class NonDominatedSorting:

    def __init__(self, epsilon=None, method="fast_non_dominated_sort") -> None:
        super().__init__()
        self.epsilon = epsilon
        self.method = method

    def do(self, F, return_rank=False, only_non_dominated_front=False, n_stop_if_ranked=None, **kwargs):
        F = F.astype(float)

        # if not set just set it to a very large values because the cython algorithms do not take None
        if n_stop_if_ranked is None:
            n_stop_if_ranked = int(1e8)
        func = load_function(self.method)

        # set the epsilon if it should be set
        if self.epsilon is not None:
            kwargs["epsilon"] = float(self.epsilon)

        fronts = func(F, **kwargs)

        # convert to numpy array for each front and filter by n_stop_if_ranked if desired
        _fronts = []
        n_ranked = 0
        for front in fronts:

            _fronts.append(np.array(front, dtype=int))

            # increment the n_ranked solution counter
            n_ranked += len(front)

            # stop if more than this solutions are n_ranked
            if n_ranked >= n_stop_if_ranked:
                break

        fronts = _fronts

        if only_non_dominated_front:
            return fronts[0]

        if return_rank:
            rank = rank_from_fronts(fronts, F.shape[0])
            return fronts, rank

        return fronts


def rank_from_fronts(fronts, n):
    # create the rank array and set values
    rank = np.full(n, 1e16, dtype=int)
    for i, front in enumerate(fronts):
        rank[front] = i

    return rank


# Returns all indices of F that are not dominated by the other objective values
def find_non_dominated(F, _F=None):
    M = Dominator.calc_domination_matrix(F, _F)
    I = np.where(np.all(M >= 0, axis=1))[0]
    return I



import numpy as np

from pymoo.core.mutation import Mutation
from pymoo.core.variable import get, Real
from pymoo.operators.crossover.binx import mut_binomial
from pymoo.operators.repair.to_bound import set_to_bounds_if_outside


# ---------------------------------------------------------------------------------------------------------
# Function
# ---------------------------------------------------------------------------------------------------------


def mut_pm(X, xl, xu, eta, prob, at_least_once):
    n, n_var = X.shape
    assert len(eta) == n
    assert len(prob) == n

    Xp = np.full(X.shape, np.inf)

    mut = mut_binomial(n, n_var, prob, at_least_once=at_least_once)
    mut[:, xl == xu] = False

    Xp[:, :] = X

    _xl = np.repeat(xl[None, :], X.shape[0], axis=0)[mut]
    _xu = np.repeat(xu[None, :], X.shape[0], axis=0)[mut]

    X = X[mut]
    eta = np.tile(eta[:, None], (1, n_var))[mut]

    delta1 = (X - _xl) / (_xu - _xl)
    delta2 = (_xu - X) / (_xu - _xl)

    mut_pow = 1.0 / (eta + 1.0)

    rand = np.random.random(X.shape)
    mask = rand <= 0.5
    mask_not = np.logical_not(mask)

    deltaq = np.zeros(X.shape)

    xy = 1.0 - delta1
    val = 2.0 * rand + (1.0 - 2.0 * rand) * (np.power(xy, (eta + 1.0)))
    d = np.power(val, mut_pow) - 1.0
    deltaq[mask] = d[mask]

    xy = 1.0 - delta2
    val = 2.0 * (1.0 - rand) + 2.0 * (rand - 0.5) * (np.power(xy, (eta + 1.0)))
    d = 1.0 - (np.power(val, mut_pow))
    deltaq[mask_not] = d[mask_not]

    # mutated values
    _Y = X + deltaq * (_xu - _xl)

    # back in bounds if necessary (floating point issues)
    _Y[_Y < _xl] = _xl[_Y < _xl]
    _Y[_Y > _xu] = _xu[_Y > _xu]

    # set the values for output
    Xp[mut] = _Y

    # in case out of bounds repair (very unlikely)
    Xp = set_to_bounds_if_outside(Xp, xl, xu)

    return Xp


# ---------------------------------------------------------------------------------------------------------
# Class
# ---------------------------------------------------------------------------------------------------------


class PolynomialMutation(Mutation):

    def __init__(self, prob=0.9, eta=20, at_least_once=False, **kwargs):
        super().__init__(prob=prob, **kwargs)
        self.at_least_once = at_least_once
        self.eta = Real(eta, bounds=(3.0, 30.0), strict=(1.0, 100.0))

    def _do(self, problem, X, params=None, **kwargs):
        X = X.astype(float)

        eta = get(self.eta, size=len(X))
        prob_var = self.get_prob_var(problem, size=len(X))

        Xp = mut_pm(X, problem.xl, problem.xu, eta, prob_var, at_least_once=self.at_least_once)

        return Xp


class PM(PolynomialMutation):
    pass




import numpy as np
from scipy.spatial.distance import cdist

from pymoo.core.indicator import Indicator
from pymoo.indicators.hv import Hypervolume
from pymoo.indicators.igd import IGD


class RMetric(Indicator):

    def __init__(self, problem, ref_points, w=None, delta=0.2, pf=None):
        """

        Parameters
        ----------
        
        problem : class
            problem instance
            
        ref_points : numpy.array
            list of reference points

        w : numpy.array
            weights for each objective

        delta : float
            The delta value representing the region of interest

        """

        Indicator.__init__(self)
        self.ref_points = ref_points
        self.problem = problem
        w_ = np.ones(self.ref_points.shape[1]) if not w else w
        self.w_points = self.ref_points + 2 * w_
        self.delta = delta

        self.pf = pf
        self.F = None
        self.others = None

    def _filter(self):

        def check_dominance(a, b, n_obj):
            flag1 = False
            flag2 = False
            for i in range(n_obj):
                if a[i] < b[i]:
                    flag1 = True
                else:
                    if a[i] > b[i]:
                        flag2 = True
            if flag1 and not flag2:
                return 1
            elif not flag1 and flag2:
                return -1
            else:
                return 0

        num_objs = np.size(self.F, axis=1)
        index_array = np.zeros(np.size(self.F, axis=0))

        # filter out all solutions that are dominated by solutions found by other algorithms
        if self.others is not None:
            for i in range(np.size(self.F, 0)):
                for j in range(np.size(self.others, 0)):
                    flag = check_dominance(self.F[i, :], self.others[j, :], num_objs)
                    if flag == -1:
                        index_array[i] = 1
                        break

        final_index = np.logical_not(index_array)
        filtered_pop = self.F[final_index, :]

        return filtered_pop

    def _preprocess(self, data, ref_point, w_point):

        datasize = np.size(data, 0)

        # Identify representative point
        ref_matrix = np.tile(ref_point, (datasize, 1))
        w_matrix = np.tile(w_point, (datasize, 1))
        # ratio of distance to the ref point over the distance between the w_point and the ref_point
        diff_matrix = (data - ref_matrix) / (w_matrix - ref_matrix)
        agg_value = np.amax(diff_matrix, axis=1)
        idx = np.argmin(agg_value)
        zp = [data[idx, :]]

        return zp,

    def _translate(self, zp, trimmed_data, ref_point, w_point):
        # Solution translation - Matlab reproduction
        # find k
        temp = ((zp[0] - ref_point) / (w_point - ref_point))
        kIdx = np.argmax(temp)

        # find zl
        temp = (zp[0][kIdx] - ref_point[kIdx]) / (w_point[kIdx] - ref_point[kIdx])
        zl = ref_point + temp * (w_point - ref_point)

        temp = zl - zp
        shift_direction = np.tile(temp, (trimmed_data.shape[0], 1))
        # new_size = self.curr_pop.shape[0]
        return trimmed_data + shift_direction

    def _trim(self, pop, centeroid, range=0.2):
        popsize, objDim = pop.shape
        diff_matrix = pop - np.tile(centeroid, (popsize, 1))[0]
        flags = np.sum(abs(diff_matrix) < range / 2, axis=1)
        filtered_matrix = pop[np.where(flags == objDim)]
        return filtered_matrix

    def _trim_fast(self, pop, centeroid, range=0.2):
        centeroid_matrix = cdist(pop, centeroid, metric='euclidean')
        filtered_matrix = pop[np.where(centeroid_matrix < range / 2), :][0]
        return filtered_matrix

    def do(self, F, others=None, calc_hv=True):
        """

        This method calculates the R-IGD and R-HV based off of the values provided.
        
        
        Parameters
        ----------

        F : numpy.ndarray
            The objective space values

        others : numpy.ndarray
            Results from other algorithms which should be used for filtering nds solutions

        calc_hv : bool
            Whether the hv is calculate - (None if more than 3 dimensions)


        Returns
        -------
        rigd : float
            R-IGD

        rhv : float
            R-HV if calc_hv is true and less or equal to 3 dimensions

        """
        self.F, self.others = F, others

        translated = []
        final_PF = []

        # 1. Prescreen Procedure - NDS Filtering
        pop = self._filter()

        pf = self.pf
        if pf is None:
            pf = self.problem.pareto_front()

        if pf is None:
            raise Exception("Please provide the Pareto front to calculate the R-Metric!")

        labels = np.argmin(cdist(pop, self.ref_points), axis=1)

        for i in range(len(self.ref_points)):
            cluster = pop[np.where(labels == i)]
            if len(cluster) != 0:
                # 2. Representative Point Identification
                zp = self._preprocess(cluster, self.ref_points[i], w_point=self.w_points[i])[0]
                # 3. Filtering Procedure - Filter points
                trimmed_data = self._trim(cluster, zp, range=self.delta)
                # 4. Solution Translation
                pop_t = self._translate(zp, trimmed_data, self.ref_points[i], w_point=self.w_points[i])
                translated.extend(pop_t)

            # 5. R-Metric Computation
            target = self._preprocess(data=pf, ref_point=self.ref_points[i], w_point=self.w_points[i])
            PF = self._trim(pf, target)
            final_PF.extend(PF)

        translated = np.array(translated)
        final_PF = np.array(final_PF)

        rigd, rhv = None, None

        if len(translated) > 0:

            # IGD Computation
            rigd = IGD(final_PF).do(translated)

            nadir_point = np.amax(self.w_points, axis=0)
            front = translated
            dim = self.ref_points[0].shape[0]
            if calc_hv:
                if dim <= 3:
                    try:
                        rhv = Hypervolume(ref_point=nadir_point).do(front)
                    except:
                        pass

        if calc_hv:
            return rigd, rhv
        else:
            return rigd



import numpy as np

from pymoo.core.individual import Individual


class Population(np.ndarray):

    def __new__(cls, individuals=[]):
        if isinstance(individuals, Individual):
            individuals = [individuals]
        return np.array(individuals).view(cls)

    def has(self, key):
        return all([ind.has(key) for ind in self])

    def collect(self, func, to_numpy=True):
        val = []
        for i in range(len(self)):
            val.append(func(self[i]))
        if to_numpy:
            val = np.array(val)
        return val

    def apply(self, func):
        self.collect(func, to_numpy=False)

    def set(self, *args, **kwargs):

        # if population is empty just return
        if self.size == 0:
            return

        # done for the old interface with the interleaving variable definition
        kwargs = interleaving_args(*args, kwargs=kwargs)

        # for each entry in the dictionary set it to each individual
        for key, values in kwargs.items():
            is_iterable = hasattr(values, '__len__') and not isinstance(values, str)

            if is_iterable and len(values) != len(self):
                raise Exception("Population Set Attribute Error: Number of values and population size do not match!")

            for i in range(len(self)):
                val = values[i] if is_iterable else values
                self[i].set(key, val)

        return self

    def get(self, *args, to_numpy=True, **kwargs):

        val = {}
        for c in args:
            val[c] = []

        # for each individual
        for i in range(len(self)):

            # for each argument
            for c in args:
                val[c].append(self[i].get(c, **kwargs))

        # convert the results to a list
        res = [val[c] for c in args]

        # to numpy array if desired - default true
        if to_numpy:
            res = [np.array(e) for e in res]

        # return as tuple or single value
        if len(args) == 1:
            return res[0]
        else:
            return tuple(res)

    @classmethod
    def merge(cls, a, b, *args):

        # do the regular merge between first and second element
        m = merge(a, b)

        # process the list of others and merge as well
        others = list(args)
        while len(others) > 0:
            m = merge(m, others.pop(0))

        return m

    @classmethod
    def create(cls, *args):
        return Population.__new__(cls, args)

    @classmethod
    def empty(cls, size=0):
        individuals = [Individual() for _ in range(size)]
        return Population.__new__(cls, individuals)

    @classmethod
    def new(cls, *args, **kwargs):
        kwargs = interleaving_args(*args, kwargs=kwargs)

        if len(kwargs) > 0:
            sizes = np.unique(np.array([len(v) for _, v in kwargs.items()]))
            if len(sizes) == 1:
                size = sizes[0]
            else:
                raise Exception(f"Population.new needs to be called with same-sized inputs, but the sizes are {sizes}")
        else:
            size = 0

        pop = Population.empty(size)
        pop.set(**kwargs)

        return pop


def pop_from_array_or_individual(array, pop=None):
    # the population type can be different - (different type of individuals)
    if pop is None:
        pop = Population.empty()

    # provide a whole population object - (individuals might be already evaluated)
    if isinstance(array, Population):
        pop = array
    elif isinstance(array, np.ndarray):
        pop = pop.new("X", np.atleast_2d(array))
    elif isinstance(array, Individual):
        pop = Population.empty(1)
        pop[0] = array
    else:
        return None

    return pop


def merge(a, b):
    if a is None:
        return b
    elif b is None:
        return a

    a, b = pop_from_array_or_individual(a), pop_from_array_or_individual(b)

    if len(a) == 0:
        return b
    elif len(b) == 0:
        return a
    else:
        obj = np.concatenate([a, b]).view(Population)
        return obj


def interleaving_args(*args, kwargs=None):
    if len(args) % 2 != 0:
        raise Exception(f"Even number of arguments are required but {len(args)} arguments were provided.")

    if kwargs is None:
        kwargs = {}

    for i in range(int(len(args) / 2)):
        key, values = args[i * 2], args[i * 2 + 1]
        kwargs[key] = values
    return kwargs


def calc_cv(pop, config=None):

    if config is None:
        config = Individual.default_config()

    G, H = pop.get("G", "H")

    from pymoo.core.individual import calc_cv as func
    CV = np.array([func(g, h, config) for g, h in zip(G, H)])
    
    return CV



import numpy as np


def func_select_by_objective(pop):
    F = pop.get("F")
    return F[:, 0].argmin()


def func_select_from_sorted(_):
    return 0


def select_by_clearing(pop, D, n_select, func_select, delta=0.05):
    clearing = EpsilonClearing(D, delta)

    while len(clearing.selected()) < n_select:
        remaining = clearing.remaining()

        if len(remaining) == 0:
            clearing.reset()
            remaining = clearing.remaining()

        best = remaining[func_select(pop[remaining])]
        clearing.select(best)

    S = clearing.selected()
    return S


class EpsilonClearing:

    def __init__(self,
                 D,
                 epsilon) -> None:

        super().__init__()

        if isinstance(D, tuple):
            self.n, self.D = D
        else:
            self.D = D
            self.n = len(D)

        self.epsilon = epsilon

        self.S = []
        self.C = np.full(self.n, False)

    def remaining(self):
        return np.where(~self.C)[0]

    def has_remaining(self):
        return self.C.sum() != self.n

    def cleared(self):
        return self.C

    def selected(self):
        return self.S

    def reset(self):
        self.C = np.full(self.n, False)
        self.C[self.S] = True

    def select(self, k):
        self.S.append(k)
        self.C[k] = True

        if callable(self.D):
            dist_to_other = self.D(k)
        else:
            dist_to_other = self.D[k]

        less_than_epsilon = dist_to_other < self.epsilon

        # problems which are currently not cleared and are supposed to
        cleared = np.where(np.logical_and(~self.C, less_than_epsilon))[0]

        # set them to be cleared
        self.C[cleared] = True

        return cleared



from pymoo.util.dominator import Dominator


def naive_non_dominated_sort(F, **kwargs):
    M = Dominator.calc_domination_matrix(F)

    fronts = []
    remaining = set(range(M.shape[0]))

    while len(remaining) > 0:

        front = []

        for i in remaining:

            is_dominated = False
            dominating = set()

            for j in front:
                rel = M[i, j]
                if rel == 1:
                    dominating.add(j)
                elif rel == -1:
                    is_dominated = True
                    break

            if is_dominated:
                continue
            else:
                front = [x for x in front if x not in dominating]
                front.append(i)

        [remaining.remove(e) for e in front]
        fronts.append(front)

    return fronts



from pymoo.docs import parse_doc_string
from pymoo.core.plot import Plot
from pymoo.util.misc import set_if_none_from_tuples
from pymoo.visualization.util import plot_circle, plot_radar_line, plot_axis_labels, equal_axis, no_ticks, \
    get_uniform_points_around_circle


class Radviz(Plot):

    def __init__(self, endpoint_style={}, **kwargs):
        """

        Radviz Plot

        Parameters
        ----------------

        axis_style : {axis_style}
        endpoint_style : dict
            Endpoints are drawn at each extreme point of an objective. This style can be modified.
        labels : {labels}

        Other Parameters
        ----------------

        figsize : {figsize}
        title : {title}
        legend : {legend}
        tight_layout : {tight_layout}
        cmap : {cmap}

        """

        super().__init__(**kwargs)

        # set the default axis style
        set_if_none_from_tuples(self.axis_style, ("color", "black"), ("linewidth", 1), ("alpha", 0.75))

        self.endpoint_style = endpoint_style
        set_if_none_from_tuples(self.endpoint_style, ("color", "black"), ("s", 70), ("alpha", 0.3))

    def _do(self):

        # initial a figure with a single plot
        self.init_figure()

        # equal axis length and no ticks
        equal_axis(self.ax)
        no_ticks(self.ax)

        V = get_uniform_points_around_circle(self.n_dim)
        plot_axis_labels(self.ax, V, self.get_labels(), **self.axis_label_style)

        # draw the outer circle and radar lines
        plot_circle(self.ax, **self.axis_style)
        plot_radar_line(self.ax, V, **self.axis_style)

        # draw the endpoints of each objective
        if self.endpoint_style:
            self.ax.scatter(V[:, 0], V[:, 1], **self.endpoint_style)

        # plot all the points
        for k, (F, kwargs) in enumerate(self.to_plot):
            N = (F[..., None] * V).sum(axis=1) / F.sum(axis=1)[:, None]
            self.ax.scatter(N[:, 0], N[:, 1], **kwargs)


parse_doc_string(Radviz.__init__)



import numpy as np
from pymoo.core.indicator import Indicator
from pymoo.indicators.distance_indicator import at_least_2d_array, derive_ideal_and_nadir_from_pf
from scipy.spatial.distance import pdist, squareform


class SpacingIndicator(Indicator):
    
    def __init__(self,
                 pf=None,
                 zero_to_one=False,
                 ideal=None,
                 nadir=None):
        """Spacing indicator
        
        The smaller the value this indicator assumes, the most uniform is the distribution of elements on the pareto front.

        Parameters
        ----------
        pf : 2d array, optional
            Pareto front, by default None
        zero_to_one : bool, optional
            Whether or not the objective values should be normalized in calculations, by default False
        ideal : 1d array, optional
            Ideal point, by default None
        nadir : 1d array, optional
            Nadir point, by default None
        """
        
        # the pareto front if necessary to calculate the indicator
        pf = at_least_2d_array(pf, extend_as="row")
        ideal, nadir = derive_ideal_and_nadir_from_pf(pf, ideal=ideal, nadir=nadir)
        
        super().__init__(pf=pf,
                         zero_to_one=zero_to_one,
                         ideal=ideal,
                         nadir=nadir)
    
    def _do(self, F, *args, **kwargs):
        
        # Get F dimensions
        n_points, n_obj = F.shape
        
        # knn
        D = squareform(pdist(F, metric="cityblock"))
        d = np.partition(D, 1, axis=1)[:, 1]
        dm = np.mean(d)
        
        # Get spacing
        S = np.sqrt(np.sum(np.square(d - dm)) / n_points)
        
        return S


from pymoo.util.ref_dirs.energy import RieszEnergyReferenceDirectionFactory
from pymoo.util.ref_dirs.energy_layer import LayerwiseRieszEnergyReferenceDirectionFactory
from pymoo.util.ref_dirs.reduction import ReductionBasedReferenceDirectionFactory
from pymoo.util.reference_direction import MultiLayerReferenceDirectionFactory


def get_reference_directions(name, *args, **kwargs):
    from pymoo.util.reference_direction import UniformReferenceDirectionFactory

    REF = {
        "uniform": UniformReferenceDirectionFactory,
        "das-dennis": UniformReferenceDirectionFactory,
        "energy": RieszEnergyReferenceDirectionFactory,
        "multi-layer": MultiLayerReferenceDirectionFactory,
        "layer-energy": LayerwiseRieszEnergyReferenceDirectionFactory,
        "reduction": ReductionBasedReferenceDirectionFactory,
    }

    if name not in REF:
        raise Exception("Reference directions factory not found.")

    return REF[name](*args, **kwargs)()



from pymoo.core.meta import Meta
from pymoo.core.problem import Problem


class StaticProblem(Meta, Problem):

    def __init__(self, problem, **kwargs):
        super().__init__(problem)
        self.kwargs = kwargs

    def _evaluate(self, _, out, *args, **kwargs):
        for K, V in self.kwargs.items():
            out[K] = V




import numpy as np

from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.core.population import Population


class DNSGA2(NSGA2):

    def __init__(self,
                 perc_detect_change=0.1,
                 perc_diversity=0.3,
                 eps=0.0,
                 version="A",
                 **kwargs):

        super().__init__(**kwargs)
        self.perc_detect_change = perc_detect_change
        self.perc_diversity = perc_diversity
        self.eps = eps
        self.version = version

    def setup(self, problem, **kwargs):
        assert not problem.has_constraints(), "DNSGA2 only works for unconstrained problems."
        return super().setup(problem, **kwargs)

    def _advance(self, **kwargs):

        pop = self.pop
        X, F = pop.get("X", "F")

        # the number of solutions to sample from the population to detect the change
        n_samples = int(np.ceil(len(pop) * self.perc_detect_change))

        # choose randomly some individuals of the current population to test if there was a change
        I = np.random.choice(np.arange(len(pop)), size=n_samples)
        samples = self.evaluator.eval(self.problem, Population.new(X=X[I]))

        # calculate the differences between the old and newly evaluated pop
        delta = ((samples.get("F") - F[I]) ** 2).mean()

        # if there is an average deviation bigger than eps -> we have a change detected
        change_detected = delta > self.eps

        if change_detected:

            # recreate the current population without being evaluated
            pop = Population.new(X=X)

            # find indices to be replaced (introduce diversity)
            I = np.where(np.random.random(len(pop)) < self.perc_diversity)[0]

            # replace with randomly sampled individuals
            if self.version == "A":
                pop[I] = self.initialization.sampling(self.problem, len(I))

            # replace by mutations of existing solutions (this occurs inplace)
            elif self.version == "B":
                self.mating.mutation(self.problem, pop[I])
            else:
                raise Exception(f"Unknown version of D-NSGA-II: {self.version}")

            # reevaluate because we know there was a change
            self.evaluator.eval(self.problem, pop)

            # do a survival to recreate rank and crowding of all individuals
            pop = self.survival.do(self.problem, pop, n_survive=len(pop))

        # create the offsprings from the current population
        off = self.mating.do(self.problem, pop, self.n_offsprings, algorithm=self)
        self.evaluator.eval(self.problem, off)

        # merge the parent population and offsprings
        pop = Population.merge(pop, off)

        # execute the survival to find the fittest solutions
        self.pop = self.survival.do(self.problem, pop, n_survive=self.pop_size, algorithm=self)




from pymoo.core.termination import NoTermination
from pymoo.termination.max_eval import MaximumFunctionCallTermination
from pymoo.util.display.single import SingleObjectiveOutput

try:
    from scipy.optimize import minimize as scipy_minimize, NonlinearConstraint, LinearConstraint
except:
    raise Exception("Please install SciPy: pip install scipy")

import warnings

import numpy as np

from pymoo.algorithms.base.local import LocalSearch
from pymoo.core.individual import Individual, constr_to_cv
from pymoo.core.population import Population

from pymoo.termination.max_gen import MaximumGenerationTermination



# ---------------------------------------------------------------------------------------------------------
# Interface
# ---------------------------------------------------------------------------------------------------------


class Optimizer(LocalSearch):

    def __init__(self, method, with_bounds=False, with_constr=False, require_jac=False,
                 use_bounds=True, use_constr=True, estm_gradients=True, disp=False, show_warnings=False, **kwargs):

        super().__init__(output=SingleObjectiveOutput(), **kwargs)

        self.method, self.with_bounds, self.with_constr, self.require_jac = method, with_bounds, with_constr, require_jac
        self.show_warnings = show_warnings
        self.use_bounds = use_bounds
        self.use_constr = use_constr
        self.estm_gradients = estm_gradients

        self.options = {
            'maxiter': int(1e8),  # because of C code interfacing this can not be inf
            'disp': disp}

    def _setup(self, problem, **kwargs):
        if isinstance(self.termination, MaximumGenerationTermination):
            self.options["maxiter"] = self.termination.n_max_gen
        elif isinstance(self.termination, MaximumFunctionCallTermination):
            self.options["maxfev"] = self.termination.n_max_evals

        self.termination = NoTermination()
        self.return_least_infeasible = True

    def _advance(self, **kwargs):
        problem, evaluator = self.problem, self.evaluator

        # add the box constraints defined in the problem
        bounds = None
        if self.use_bounds:

            xl, xu = self.problem.bounds()
            if self.with_bounds:
                bounds = np.column_stack([xl, xu])
            else:
                if xl is not None or xu is not None:
                    raise Exception(f"Error: Boundary constraints can not be handled by {self.method}")

        # define the actual constraints if supported by the algorithm
        constr = []
        if self.use_constr:

            constr = [LinearConstraint(np.eye(self.problem.n_var), xl, xu)]

            if problem.has_constraints():

                if self.with_constr:
                    def fun_constr(x):
                        g = problem.evaluate(x, return_values_of=["G"])
                        cv = constr_to_cv(g)
                        return cv

                    non_lin_constr = NonlinearConstraint(fun_constr, -float("inf"), 0)

                    constr.append(non_lin_constr)

                else:
                    raise Exception(f"Error: Constraint handling is not supported by {self.method}")

        # the objective function to be optimized and add gradients if available
        if self.estm_gradients:
            jac = None

            def fun_obj(x):
                f = problem.evaluate(x, return_values_of=["F"])[0]
                evaluator.n_eval += 1
                return f

        else:
            jac = True

            def fun_obj(x):
                f, df = problem.evaluate(x, return_values_of=["F", "dF"])

                if df is None:
                    raise Exception("If the gradient shall not be estimate, please set out['dF'] in _evaluate. ")

                evaluator.n_eval += 1
                return f[0], df[0]

        # the arguments to be used
        kwargs = dict(args=(), method=self.method, bounds=bounds, constraints=constr, jac=jac, options=self.options)

        # the starting solution found by sampling beforehand
        x0 = self.opt[0].X

        # actually run the optimization
        if not self.show_warnings:
            warnings.simplefilter("ignore")

        res = scipy_minimize(fun_obj, x0, **kwargs)

        opt = Population.create(Individual(X=res.x))
        self.evaluator.eval(self.problem, opt, algorithm=self)

        self.pop, self.off = opt, opt

        self.termination.force_termination = True

        if hasattr("res", "nit"):
            self.n_gen = res.nit + 1


# ---------------------------------------------------------------------------------------------------------
# Object Oriented Interface
# ---------------------------------------------------------------------------------------------------------

# +++++++++++++++++++++++++++++++++++++++++
# UNCONSTRAINED
# +++++++++++++++++++++++++++++++++++++++++

class NelderMead(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("Nelder-Mead", **kwargs)


class CG(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("CG", require_jac=True, **kwargs)


class NewtonCG(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("Newton-CG", require_jac=True, **kwargs)


class BFGS(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("BFGS", **kwargs)


class Powell(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("Powell", **kwargs)


class Dogleg(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("dogleg", require_jac=True, **kwargs)


class TrustNCG(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("trust-ncg", require_jac=True, **kwargs)


class TrustExact(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("trust-exact", require_jac=True, **kwargs)


class TrustKrylov(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("trust-krylov", require_jac=True, **kwargs)


# +++++++++++++++++++++++++++++++++++++++++
# BOX CONSTRAINS
# +++++++++++++++++++++++++++++++++++++++++


class LBFGSB(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("L-BFGS-B", with_bounds=True, **kwargs)


class TNC(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("TNC", with_bounds=True, **kwargs)


# +++++++++++++++++++++++++++++++++++++++++
# NON-LINEAR CONSTRAINTS
# +++++++++++++++++++++++++++++++++++++++++


class COBYLA(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("COBYLA", with_bounds=False, with_constr=True, **kwargs)


class SLSQP(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("SLSQP", with_bounds=True, with_constr=True, **kwargs)


class TrustConstr(Optimizer):

    def __init__(self, **kwargs):
        super().__init__("trust-constr", with_bounds=True, with_constr=True, **kwargs)



from pymoo.core.algorithm import Algorithm
from pymoo.core.population import Population
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.util.display.single import SingleObjectiveOutput


class RandomSearch(Algorithm):

    def __init__(self,
                 n_points_per_iteration=100,
                 sampling=FloatRandomSampling(),
                 output=SingleObjectiveOutput(),
                 **kwargs):
        super().__init__(output=output, **kwargs)
        self.n_points_per_iteration = n_points_per_iteration
        self.sampling = sampling

    def _initialize_infill(self):
        return self._infill()

    def _infill(self):
        return self.sampling.do(self.problem, self.n_points_per_iteration)

    def _advance(self, infills=None, **kwargs):
        self.pop = infills if self.opt is None else Population.merge(infills, self.opt)



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.util.misc import crossover_mask


class PointCrossover(Crossover):

    def __init__(self, n_points, **kwargs):
        super().__init__(2, 2, **kwargs)
        self.n_points = n_points

    def _do(self, _, X, **kwargs):

        # get the X of parents and count the matings
        _, n_matings, n_var = X.shape

        # start point of crossover
        r = np.row_stack([np.random.permutation(n_var - 1) + 1 for _ in range(n_matings)])[:, :self.n_points]
        r.sort(axis=1)
        r = np.column_stack([r, np.full(n_matings, n_var)])

        # the mask do to the crossover
        M = np.full((n_matings, n_var), False)

        # create for each individual the crossover range
        for i in range(n_matings):

            j = 0
            while j < r.shape[1] - 1:
                a, b = r[i, j], r[i, j + 1]
                M[i, a:b] = True
                j += 2

        Xp = crossover_mask(X, M)

        return Xp


class SinglePointCrossover(PointCrossover):

    def __init__(self, **kwargs):
        super().__init__(n_points=1, **kwargs)


class TwoPointCrossover(PointCrossover):

    def __init__(self, **kwargs):
        super().__init__(n_points=2, **kwargs)



import numpy as np

from pymoo.core.meta import Meta
from pymoo.core.problem import Problem
from pymoo.util.normalization import denormalize, normalize


class ZeroToOne(Meta, Problem):

    def __init__(self, problem):
        super().__init__(problem)
        assert self.xl is not None and self.xu is not None, "Both, xl and xu, must be set to redefine the problem!"
        self.xl, self.xu = np.zeros(self.n_var), np.ones(self.n_var)

    def do(self, X, return_values_of, *args, **kwargs):
        Xp = self.normalize(X)

        out = super().do(Xp, return_values_of, *args, **kwargs)
        out["__X__"] = X

        return out

    def denormalize(self, x):
        return denormalize(x, self.__object__.xl, self.__object__.xu)

    def normalize(self, x):
        return normalize(x, self.__object__.xl, self.__object__.xu)



import numpy as np

from pymoo.core.operator import Operator
from pymoo.core.population import Population
from pymoo.core.variable import Real, get


class Crossover(Operator):

    def __init__(self,
                 n_parents,
                 n_offsprings,
                 prob=0.9,
                 **kwargs):
        super().__init__(**kwargs)
        self.n_parents = n_parents
        self.n_offsprings = n_offsprings
        self.prob = Real(prob, bounds=(0.5, 1.0), strict=(0.0, 1.0))

    def do(self, problem, pop, parents=None, **kwargs):

        # if a parents with array with mating indices is provided -> transform the input first
        if parents is not None:
            pop = [pop[mating] for mating in parents]

        # get the dimensions necessary to create in and output
        n_parents, n_offsprings = self.n_parents, self.n_offsprings
        n_matings, n_var = len(pop), problem.n_var

        # get the actual values from each of the parents
        X = np.swapaxes(np.array([[parent.get("X") for parent in mating] for mating in pop]), 0, 1)
        if self.vtype is not None:
            X = X.astype(self.vtype)

        # the array where the offsprings will be stored to
        Xp = np.empty(shape=(n_offsprings, n_matings, n_var), dtype=X.dtype)

        # the probability of executing the crossover
        prob = get(self.prob, size=n_matings)

        # a boolean mask when crossover is actually executed
        cross = np.random.random(n_matings) < prob

        # the design space from the parents used for the crossover
        if np.any(cross):

            # we can not prefilter for cross first, because there might be other variables using the same shape as X
            Q = self._do(problem, X, **kwargs)
            assert Q.shape == (n_offsprings, n_matings, problem.n_var), "Shape is incorrect of crossover impl."
            Xp[:, cross] = Q[:, cross]

        # now set the parents whenever NO crossover has been applied
        for k in np.flatnonzero(~cross):
            if n_offsprings < n_parents:
                s = np.random.choice(np.arange(self.n_parents), size=n_offsprings, replace=False)
            elif n_offsprings == n_parents:
                s = np.arange(n_parents)
            else:
                s = []
                while len(s) < n_offsprings:
                    s.extend(np.random.permutation(n_parents))
                s = s[:n_offsprings]

            Xp[:, k] = np.copy(X[s, k])

        # flatten the array to become a 2d-array
        Xp = Xp.reshape(-1, X.shape[-1])

        # create a population object
        off = Population.new("X", Xp)

        return off

    def _do(self, problem, X, **kwargs):
        pass





from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.algorithm import Algorithm, LoopwiseAlgorithm
from pymoo.core.initialization import Initialization
from pymoo.operators.sampling.lhs import LatinHypercubeSampling
from pymoo.termination.ftol import SingleObjectiveSpaceTermination
from pymoo.termination.robust import RobustTermination
from pymoo.util.display.single import SingleObjectiveOutput


class LocalSearch(LoopwiseAlgorithm, Algorithm):

    def __init__(self,
                 initial=None,
                 sampling=LatinHypercubeSampling(),
                 output=SingleObjectiveOutput(),
                 n_sample_points=20,
                 **kwargs):
        super().__init__(output=output, **kwargs)

        # the default termination if not specified otherwise
        self.termination = RobustTermination(SingleObjectiveSpaceTermination(tol=1e-8), period=10)

        # the type of initial sampling
        initial = initial if "x0" not in kwargs else kwargs["x0"]
        if initial is not None:
            sampling = initial

        self.initialization = Initialization(sampling)

        # the number of sampling points to determine x0
        self.n_sample_points = n_sample_points

    def _initialize_infill(self):
        return self.initialization.do(self.problem, self.n_sample_points, algorithm=self)

    def _initialize_advance(self, infills=None, **kwargs):
        self.x0 = FitnessSurvival().do(self.problem, infills, n_survive=1, algorithm=self)[0]





import numpy as np

from pymoo.util.misc import cdist


def default_attr(pop):
    return pop.get("X")


class DuplicateElimination:

    def __init__(self, func=None) -> None:
        super().__init__()
        self.func = func

        if self.func is None:
            self.func = default_attr

    def do(self, pop, *args, return_indices=False, to_itself=True):
        original = pop

        if len(pop) == 0:
            return (pop, [], []) if return_indices else pop

        if to_itself:
            pop = pop[~self._do(pop, None, np.full(len(pop), False))]

        for arg in args:
            if len(arg) > 0:

                if len(pop) == 0:
                    break
                elif len(arg) == 0:
                    continue
                else:
                    pop = pop[~self._do(pop, arg, np.full(len(pop), False))]

        if return_indices:
            no_duplicate, is_duplicate = [], []
            H = set(pop)

            for i, ind in enumerate(original):
                if ind in H:
                    no_duplicate.append(i)
                else:
                    is_duplicate.append(i)

            return pop, no_duplicate, is_duplicate
        else:
            return pop

    def _do(self, pop, other, is_duplicate):
        return is_duplicate


class DefaultDuplicateElimination(DuplicateElimination):

    def __init__(self, epsilon=1e-16, **kwargs) -> None:
        super().__init__(**kwargs)
        self.epsilon = epsilon

    def calc_dist(self, pop, other=None):
        X = self.func(pop)

        if other is None:
            D = cdist(X, X)
            D[np.triu_indices(len(X))] = np.inf
        else:
            _X = self.func(other)
            D = cdist(X, _X)

        return D

    def _do(self, pop, other, is_duplicate):
        D = self.calc_dist(pop, other)
        D[np.isnan(D)] = np.inf

        is_duplicate[np.any(D <= self.epsilon, axis=1)] = True
        return is_duplicate


def to_float(val):
    if isinstance(val, bool) or isinstance(val, np.bool_):
        return 0.0 if val else 1.0
    else:
        return val


class ElementwiseDuplicateElimination(DefaultDuplicateElimination):

    def __init__(self, cmp_func=None, **kwargs) -> None:
        super().__init__(**kwargs)

        if cmp_func is None:
            cmp_func = self.is_equal

        self.cmp_func = cmp_func

    def is_equal(self, a, b):
        pass

    def _do(self, pop, other, is_duplicate):

        if other is None:
            for i in range(len(pop)):
                for j in range(i + 1, len(pop)):
                    val = to_float(self.cmp_func(pop[i], pop[j]))
                    if val < self.epsilon:
                        is_duplicate[i] = True
                        break
        else:
            for i in range(len(pop)):
                for j in range(len(other)):
                    val = to_float(self.cmp_func(pop[i], other[j]))
                    if val < self.epsilon:
                        is_duplicate[i] = True
                        break

        return is_duplicate


def to_hash(x):
    try:
        h = hash(x)
    except:
        try:
            h = hash(str(x))
        except:
            raise Exception("Hash could not be calculated. Please use another duplicate elimination.")

    return h


class HashDuplicateElimination(DuplicateElimination):

    def __init__(self, func=to_hash) -> None:
        super().__init__()
        self.func = func

    def _do(self, pop, other, is_duplicate):
        H = set()

        if other is not None:
            for o in other:
                val = self.func(o)
                H.add(self.func(val))

        for i, ind in enumerate(pop):
            val = self.func(ind)
            h = self.func(val)

            if h in H:
                is_duplicate[i] = True
            else:
                H.add(h)

        return is_duplicate


class NoDuplicateElimination(DuplicateElimination):

    def do(self, pop, *args, **kwargs):
        return pop



from pymoo.algorithms.base.bracket import BracketSearch
from pymoo.core.individual import Individual
from pymoo.core.population import Population


def quadr_interp_equ(xa, fa, xb, fb, xc, fc):
    g1 = (fc - fa) / (xc - xa)
    g2 = ((fb - fc) / (xb - xc) - g1) * (xb - xa)
    xd = 0.5 * ((xa - xb) - g1 / g2)
    return xd


def quadr_interp(a, b, c):
    return Individual(X=quadr_interp_equ(a.X, a.F, b.X, b.F, c.X, c.F))


class QuadraticInterpolationSearch(BracketSearch):

    def __init__(self, a=None, b=None, **kwargs):
        """

        7.1.2 Quadratic Interpolation Search
        http://www.mathcs.emory.edu/~haber/math315/chap7.pdf

        Parameters
        ----------
        a
        b
        kwargs
        """
        super().__init__(a, b, **kwargs)

    def _initialize_infill(self):
        super()._initialize_infill()
        a, b = self.a, self.b

        # set c to be directly in the middle between the two brackets
        c = Individual(X=(b.X - a.X) / 2)

        # create a population with all three individuals
        pop = Population.create(a, b, c)

        return pop

    def _advance(self, **kwargs):

        # all the elements in the interval
        a, b, c = self.pop

        # if this is the case then the function is not convex (which means U shaped)
        if c.F[0] >= a.F[0] or c.F[0] >= b.F[0]:

            # choose the left side if a smaller than b, or the right side otherwise
            if a.F[0] <= b.F[0]:
                a = c
            else:
                b = c

            c = Individual(X=(b.X - a.X) / 2)
            self.evaluator.eval(self.problem, c, algorithm=self)
            self.infills = c

        else:

            d = quadr_interp(a, b, c)
            self.evaluator.eval(self.problem, d, algorithm=self)
            self.infills = d

            # swap c and d -> make sure d is always on the right of c -> a, c, d, b
            if c.X[0] > d.X[0]:
                c, d = d, c

            # if c is better than d, then d becomes the new right bound
            if c.F[0] <= d.F[0]:
                b = d

            # if d is better than c, then c becomes the new left bound and d the new right bound
            else:
                a, c = c, d

        self.pop = Population.create(a, b, c)




def get_problem(name, *args, **kwargs):
    name = name.lower()

    if name.startswith("bbob-"):
        from pymoo.vendor.vendor_coco import COCOProblem
        return COCOProblem(name.lower(), **kwargs)

    from pymoo.problems.multi import BNH, Carside
    from pymoo.problems.multi import CTP1, CTP2, CTP3, CTP4, CTP5, CTP6, CTP7, CTP8
    from pymoo.problems.multi import DASCMOP1, DASCMOP2, DASCMOP3, DASCMOP4, DASCMOP5, DASCMOP6, DASCMOP7, DASCMOP8, \
        DASCMOP9
    from pymoo.problems.multi import MODAct, MW1, MW2, MW3, MW4, MW5, MW6, MW7, MW8, MW9, MW10, MW11, MW12, MW13, MW14
    from pymoo.problems.single import Ackley
    from pymoo.problems.many import DTLZ1, C1DTLZ1, DC1DTLZ1, DC1DTLZ3, DC2DTLZ1, DC2DTLZ3, DC3DTLZ1, DC3DTLZ3, C1DTLZ3, \
        C2DTLZ2, C3DTLZ1, C3DTLZ4, ScaledDTLZ1, ConvexDTLZ2, ConvexDTLZ4, DTLZ2, DTLZ3, DTLZ4, DTLZ5, DTLZ6, DTLZ7, \
        InvertedDTLZ1, WFG1, WFG2, WFG3, WFG4, WFG5, WFG6, WFG7, WFG8, WFG9
    from pymoo.problems.multi import Kursawe, OSY, SRN, TNK, Truss2D, WeldedBeam, ZDT1, ZDT2, ZDT3, ZDT4, ZDT5, ZDT6
    from pymoo.problems.single import CantileveredBeam, Griewank, Himmelblau, Knapsack, PressureVessel, Rastrigin, \
        Rosenbrock, Schwefel, Sphere, Zakharov
    from pymoo.problems.single import G1, G2, G3, G4, G5, G6, G7, G8, G9, G10, G11, G12, G13, G14, G15, G16, G17, G18, \
        G19, G20, G21, G22, G23, G24
    from pymoo.problems.dynamic.df import DF1, DF2, DF3, DF4, DF5, DF6, DF7, DF8, DF9, DF10, DF11, DF12, DF13, DF14

    PROBLEM = {
        'ackley': Ackley,
        'bnh': BNH,
        'carside': Carside,
        'ctp1': CTP1,
        'ctp2': CTP2,
        'ctp3': CTP3,
        'ctp4': CTP4,
        'ctp5': CTP5,
        'ctp6': CTP6,
        'ctp7': CTP7,
        'ctp8': CTP8,
        'dascmop1': DASCMOP1,
        'dascmop2': DASCMOP2,
        'dascmop3': DASCMOP3,
        'dascmop4': DASCMOP4,
        'dascmop5': DASCMOP5,
        'dascmop6': DASCMOP6,
        'dascmop7': DASCMOP7,
        'dascmop8': DASCMOP8,
        'dascmop9': DASCMOP9,
        'df1': DF1,
        'df2': DF2,
        'df3': DF3,
        'df4': DF4,
        'df5': DF5,
        'df6': DF6,
        'df7': DF7,
        'df8': DF8,
        'df9': DF9,
        'df10': DF10,
        'df11': DF11,
        'df12': DF12,
        'df13': DF13,
        'df14': DF14,
        'modact': MODAct,
        'mw1': MW1,
        'mw2': MW2,
        'mw3': MW3,
        'mw4': MW4,
        'mw5': MW5,
        'mw6': MW6,
        'mw7': MW7,
        'mw8': MW8,
        'mw9': MW9,
        'mw10': MW10,
        'mw11': MW11,
        'mw12': MW12,
        'mw13': MW13,
        'mw14': MW14,
        'dtlz1^-1': InvertedDTLZ1,
        'dtlz1': DTLZ1,
        'dtlz2': DTLZ2,
        'dtlz3': DTLZ3,
        'dtlz4': DTLZ4,
        'dtlz5': DTLZ5,
        'dtlz6': DTLZ6,
        'dtlz7': DTLZ7,
        'convex_dtlz2': ConvexDTLZ2,
        'convex_dtlz4': ConvexDTLZ4,
        'sdtlz1': ScaledDTLZ1,
        'c1dtlz1': C1DTLZ1,
        'c1dtlz3': C1DTLZ3,
        'c2dtlz2': C2DTLZ2,
        'c3dtlz1': C3DTLZ1,
        'c3dtlz4': C3DTLZ4,
        'dc1dtlz1': DC1DTLZ1,
        'dc1dtlz3': DC1DTLZ3,
        'dc2dtlz1': DC2DTLZ1,
        'dc2dtlz3': DC2DTLZ3,
        'dc3dtlz1': DC3DTLZ1,
        'dc3dtlz3': DC3DTLZ3,
        'cantilevered_beam': CantileveredBeam,
        'griewank': Griewank,
        'himmelblau': Himmelblau,
        'knp': Knapsack,
        'kursawe': Kursawe,
        'osy': OSY,
        'pressure_vessel': PressureVessel,
        'rastrigin': Rastrigin,
        'rosenbrock': Rosenbrock,
        'schwefel': Schwefel,
        'sphere': Sphere,
        'srn': SRN,
        'tnk': TNK,
        'truss2d': Truss2D,
        'welded_beam': WeldedBeam,
        'zakharov': Zakharov,
        'zdt1': ZDT1,
        'zdt2': ZDT2,
        'zdt3': ZDT3,
        'zdt4': ZDT4,
        'zdt5': ZDT5,
        'zdt6': ZDT6,
        'g1': G1,
        'g2': G2,
        'g3': G3,
        'g4': G4,
        'g5': G5,
        'g6': G6,
        'g7': G7,
        'g8': G8,
        'g9': G9,
        'g10': G10,
        'g11': G11,
        'g12': G12,
        'g13': G13,
        'g14': G14,
        'g15': G15,
        'g16': G16,
        'g17': G17,
        'g18': G18,
        'g19': G19,
        'g20': G20,
        'g21': G21,
        'g22': G22,
        'g23': G23,
        'g24': G24,
        'wfg1': WFG1,
        'wfg2': WFG2,
        'wfg3': WFG3,
        'wfg4': WFG4,
        'wfg5': WFG5,
        'wfg6': WFG6,
        'wfg7': WFG7,
        'wfg8': WFG8,
        'wfg9': WFG9
    }

    if name not in PROBLEM:
        raise Exception("Problem not found.")

    return PROBLEM[name](*args, **kwargs)



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.core.population import Population


class NoCrossover(Crossover):

    def __init__(self):
        super().__init__(1, 1, 0.0)

    def do(self, problem, pop, **kwargs):
        return Population.create(*[np.random.choice(parents) for parents in pop])



import warnings

import numpy as np
from numpy.linalg import LinAlgError

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import TournamentSelection, compare
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.function_loader import load_function
from pymoo.util.misc import intersect, has_feasible
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


# =========================================================================================================
# Implementation
# =========================================================================================================

def comp_by_cv_then_random(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop[a].CV > 0.0 or pop[b].CV > 0.0:
            S[i] = compare(a, pop[a].CV, b, pop[b].CV, method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible just set random
        else:
            S[i] = np.random.choice([a, b])

    return S[:, None].astype(int)


class NSGA3(GeneticAlgorithm):

    def __init__(self,
                 ref_dirs,
                 pop_size=None,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(func_comp=comp_by_cv_then_random),
                 crossover=SBX(eta=30, prob=1.0),
                 mutation=PM(eta=20),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 output=MultiObjectiveOutput(),
                 **kwargs):
        """

        Parameters
        ----------

        ref_dirs : {ref_dirs}
        pop_size : int (default = None)
            By default the population size is set to None which means that it will be equal to the number of reference
            line. However, if desired this can be overwritten by providing a positive number.
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        n_offsprings : {n_offsprings}

        """

        self.ref_dirs = ref_dirs

        # in case of R-NSGA-3 they will be None - otherwise this will be executed
        if self.ref_dirs is not None:

            if pop_size is None:
                pop_size = len(self.ref_dirs)

            if pop_size < len(self.ref_dirs):
                print(
                    f"WARNING: pop_size={pop_size} is less than the number of reference directions ref_dirs={len(self.ref_dirs)}.\n"
                    "This might cause unwanted behavior of the algorithm. \n"
                    "Please make sure pop_size is equal or larger than the number of reference directions. ")

        if 'survival' in kwargs:
            survival = kwargs['survival']
            del kwargs['survival']
        else:
            survival = ReferenceDirectionSurvival(ref_dirs)

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         output=output,
                         advance_after_initial_infill=True,
                         **kwargs)

    def _setup(self, problem, **kwargs):

        if self.ref_dirs is not None:
            if self.ref_dirs.shape[1] != problem.n_obj:
                raise Exception(
                    "Dimensionality of reference points must be equal to the number of objectives: %s != %s" %
                    (self.ref_dirs.shape[1], problem.n_obj))

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.survival.opt


# =========================================================================================================
# Survival
# =========================================================================================================


class ReferenceDirectionSurvival(Survival):

    def __init__(self, ref_dirs):
        super().__init__(filter_infeasible=True)
        self.ref_dirs = ref_dirs
        self.opt = None
        self.norm = HyperplaneNormalization(ref_dirs.shape[1])

    def _do(self, problem, pop, n_survive, D=None, **kwargs):

        # attributes to be set after the survival
        F = pop.get("F")

        # calculate the fronts of the population
        fronts, rank = NonDominatedSorting().do(F, return_rank=True, n_stop_if_ranked=n_survive)
        non_dominated, last_front = fronts[0], fronts[-1]

        # update the hyperplane based boundary estimation
        hyp_norm = self.norm
        hyp_norm.update(F, nds=non_dominated)
        ideal, nadir = hyp_norm.ideal_point, hyp_norm.nadir_point

        #  consider only the population until we come to the splitting front
        I = np.concatenate(fronts)
        pop, rank, F = pop[I], rank[I], F[I]

        # update the front indices for the current population
        counter = 0
        for i in range(len(fronts)):
            for j in range(len(fronts[i])):
                fronts[i][j] = counter
                counter += 1
        last_front = fronts[-1]

        # associate individuals to niches
        niche_of_individuals, dist_to_niche, dist_matrix = \
            associate_to_niches(F, self.ref_dirs, ideal, nadir)

        # attributes of a population
        pop.set('rank', rank,
                'niche', niche_of_individuals,
                'dist_to_niche', dist_to_niche)

        # set the optimum, first front and closest to all reference directions
        closest = np.unique(dist_matrix[:, np.unique(niche_of_individuals)].argmin(axis=0))
        self.opt = pop[intersect(fronts[0], closest)]

        # if we need to select individuals to survive
        if len(pop) > n_survive:

            # if there is only one front
            if len(fronts) == 1:
                n_remaining = n_survive
                until_last_front = np.array([], dtype=int)
                niche_count = np.zeros(len(self.ref_dirs), dtype=int)

            # if some individuals already survived
            else:
                until_last_front = np.concatenate(fronts[:-1])
                niche_count = calc_niche_count(len(self.ref_dirs), niche_of_individuals[until_last_front])
                n_remaining = n_survive - len(until_last_front)

            S = niching(pop[last_front], n_remaining, niche_count, niche_of_individuals[last_front],
                        dist_to_niche[last_front])

            survivors = np.concatenate((until_last_front, last_front[S].tolist()))
            pop = pop[survivors]

        return pop


def niching(pop, n_remaining, niche_count, niche_of_individuals, dist_to_niche):
    survivors = []

    # boolean array of elements that are considered for each iteration
    mask = np.full(len(pop), True)

    while len(survivors) < n_remaining:

        # number of individuals to select in this iteration
        n_select = n_remaining - len(survivors)

        # all niches where new individuals can be assigned to and the corresponding niche count
        next_niches_list = np.unique(niche_of_individuals[mask])
        next_niche_count = niche_count[next_niches_list]

        # the minimum niche count
        min_niche_count = next_niche_count.min()

        # all niches with the minimum niche count (truncate if randomly if more niches than remaining individuals)
        next_niches = next_niches_list[np.where(next_niche_count == min_niche_count)[0]]
        next_niches = next_niches[np.random.permutation(len(next_niches))[:n_select]]

        for next_niche in next_niches:

            # indices of individuals that are considered and assign to next_niche
            next_ind = np.where(np.logical_and(niche_of_individuals == next_niche, mask))[0]

            # shuffle to break random tie (equal perp. dist) or select randomly
            np.random.shuffle(next_ind)

            if niche_count[next_niche] == 0:
                next_ind = next_ind[np.argmin(dist_to_niche[next_ind])]
            else:
                # already randomized through shuffling
                next_ind = next_ind[0]

            # add the selected individual to the survivors
            mask[next_ind] = False
            survivors.append(int(next_ind))

            # increase the corresponding niche count
            niche_count[next_niche] += 1

    return survivors


def associate_to_niches(F, niches, ideal_point, nadir_point, utopian_epsilon=0.0):
    utopian_point = ideal_point - utopian_epsilon

    denom = nadir_point - utopian_point
    denom[denom == 0] = 1e-12

    # normalize by ideal point and intercepts
    N = (F - utopian_point) / denom
    dist_matrix = load_function("calc_perpendicular_distance")(N, niches)

    niche_of_individuals = np.argmin(dist_matrix, axis=1)
    dist_to_niche = dist_matrix[np.arange(F.shape[0]), niche_of_individuals]

    return niche_of_individuals, dist_to_niche, dist_matrix


def calc_niche_count(n_niches, niche_of_individuals):
    niche_count = np.zeros(n_niches, dtype=int)
    index, count = np.unique(niche_of_individuals, return_counts=True)
    niche_count[index] = count
    return niche_count


# =========================================================================================================
# Normalization
# =========================================================================================================


class HyperplaneNormalization:

    def __init__(self, n_dim) -> None:
        super().__init__()
        self.ideal_point = np.full(n_dim, np.inf)
        self.worst_point = np.full(n_dim, -np.inf)
        self.nadir_point = None
        self.extreme_points = None

    def update(self, F, nds=None):

        # find or usually update the new ideal point - from feasible solutions
        self.ideal_point = np.min(np.vstack((self.ideal_point, F)), axis=0)
        self.worst_point = np.max(np.vstack((self.worst_point, F)), axis=0)

        # this decides whether only non-dominated points or all points are used to determine the extreme points
        if nds is None:
            nds = np.arange(len(F))

        # find the extreme points for normalization
        self.extreme_points = get_extreme_points_c(F[nds, :], self.ideal_point,
                                                   extreme_points=self.extreme_points)

        # find the intercepts for normalization and do backup if gaussian elimination fails
        worst_of_population = np.max(F, axis=0)
        worst_of_front = np.max(F[nds, :], axis=0)

        self.nadir_point = get_nadir_point(self.extreme_points, self.ideal_point, self.worst_point,
                                           worst_of_population, worst_of_front)


def get_extreme_points_c(F, ideal_point, extreme_points=None):
    # calculate the asf which is used for the extreme point decomposition
    weights = np.eye(F.shape[1])
    weights[weights == 0] = 1e6

    # add the old extreme points to never loose them for normalization
    _F = F
    if extreme_points is not None:
        _F = np.concatenate([extreme_points, _F], axis=0)

    # use __F because we substitute small values to be 0
    __F = _F - ideal_point
    __F[__F < 1e-3] = 0

    # update the extreme points for the normalization having the highest asf value each
    F_asf = np.max(__F * weights[:, None, :], axis=2)

    I = np.argmin(F_asf, axis=1)
    extreme_points = _F[I, :]

    return extreme_points


def get_nadir_point(extreme_points, ideal_point, worst_point, worst_of_front, worst_of_population):
    try:

        # find the intercepts using gaussian elimination
        M = extreme_points - ideal_point
        b = np.ones(extreme_points.shape[1])
        plane = np.linalg.solve(M, b)

        warnings.simplefilter("ignore")
        intercepts = 1 / plane

        nadir_point = ideal_point + intercepts

        # check if the hyperplane makes sense
        if not np.allclose(np.dot(M, plane), b) or np.any(intercepts <= 1e-6):
            raise LinAlgError()

        # if the nadir point should be larger than any value discovered so far set it to that value
        # NOTE: different to the proposed version in the paper
        b = nadir_point > worst_point
        nadir_point[b] = worst_point[b]

    except LinAlgError:

        # fall back to worst of front otherwise
        nadir_point = worst_of_front

    # if the range is too small set it to worst of population
    b = nadir_point - ideal_point <= 1e-6
    nadir_point[b] = worst_of_population[b]

    return nadir_point


parse_doc_string(NSGA3.__init__)



import sys

import numpy as np
from matplotlib import patches
from matplotlib.collections import PatchCollection


def get_circle_points(n_points):
    t = np.linspace(0, 2 * np.pi, n_points, endpoint=False)
    return np.column_stack([np.cos(t), np.sin(t)])


def default_number_to_text(val):
    if val > 1e3:
        return "{:.2e}".format(val)
    else:
        return "{:.2f}".format(val)


def in_notebook():
    return 'ipykernel' in sys.modules


def get_uniform_points_around_circle(n):
    t = 2 * np.pi * np.arange(n) / n
    s = np.column_stack([np.cos(t), np.sin(t)])
    return s


def plot_circle(ax, center=0, radius=1, **kwargs):
    P = get_circle_points(5000)
    P = (P + center) * radius
    ax.plot(P[:, 0], P[:, 1], **kwargs)


def plot_radar_line(ax, x, **kwargs):
    x = np.row_stack([x, x[0]])
    ax.plot(x[:, 0], x[:, 1], **kwargs)


def plot_axes_arrow(ax, X, extend_factor=1.0, **kwargs):
    for (x, y) in X:
        ax.arrow(0, 0, x * extend_factor, y * extend_factor, **kwargs)


def plot_axes_lines(ax, X, extend_factor=1.0, **kwargs):
    for (x, y) in X:
        ax.plot([0, x * extend_factor], [0, y * extend_factor], **kwargs)


def plot_polygon(ax, x, **kwargs):
    ax.add_collection(PatchCollection([patches.Polygon(x, True)], **kwargs))


def plot_axis_labels(ax, endpoints, labels, margin=0.035, size='small', **kwargs):
    for k in range(len(labels)):
        xy = endpoints[k]

        if xy[0] < 0.0:
            x = xy[0] - margin
            ha = "right"
        else:
            x = xy[0] + margin
            ha = "left"

        if xy[1] < 0.0:
            y = xy[1] - margin
            va = "top"
        else:
            y = xy[1] + margin
            va = "bottom"

        ax.text(x, y, labels[k], ha=ha, va=va, size=size, **kwargs)


def equal_axis(ax):
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.axis('equal')


def no_ticks(ax):
    ax.set_yticks([])
    ax.set_xticks([])
    ax.set_frame_on(False)


def normalize(data, bounds, reverse=False, return_bounds=False):
    from pymoo.util.normalization import normalize as _normalize

    _F = np.row_stack([e[0] for e in data])
    if bounds is None:
        bounds = (_F.min(axis=0), _F.max(axis=0))

    to_plot = []
    for k in range(len(data)):
        F = _normalize(data[k][0], bounds[0], bounds[1])

        if reverse:
            F = 1 - F

        to_plot.append([F, data[k][1]])

    if return_bounds:
        return to_plot, bounds
    else:
        return to_plot


def parse_bounds(bounds, n_dim):
    if bounds is not None:
        bounds = np.array(bounds, dtype=float)
        if bounds.ndim == 1:
            bounds = bounds[None, :].repeat(n_dim, axis=0).T
    return bounds


def radviz_pandas(F):
    import pandas as pd
    df = pd.DataFrame([x for x in F], columns=["X%s" % k for k in range(F.shape[1])])
    df["class"] = "Points"
    return pd.plotting.radviz(df, "class")




from pymoo.algorithms.moo.nsga2 import RankAndCrowdingSurvival
from pymoo.algorithms.soo.nonconvex.de import DE, Variant
from pymoo.core.population import Population

from pymoo.docs import parse_doc_string
from pymoo.operators.control import NoParameterControl
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.dominator import get_relation
from pymoo.termination.default import DefaultMultiObjectiveTermination


class GDE3(DE):

    def __init__(self, variant=None, **kwargs):

        if variant is None:

            if "control" not in kwargs:
                kwargs["control"] = NoParameterControl

            # the setting as proposed in the paper
            variant = Variant(selection="rand", F=0.1, CR=0.9, **kwargs)

        super().__init__(variant=variant, output=MultiObjectiveOutput(), **kwargs)
        self.termination = DefaultMultiObjectiveTermination()

    def _initialize_advance(self, infills=None, **kwargs):
        RankAndCrowdingSurvival().do(self.problem, infills, return_indices=True)

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus 'infills' must to be provided."
        pop = self.pop

        # the pool of solutions considered to survive
        pool = pop.tolist()

        # now do the replacement of individuals
        for infill in infills:
            k = infill.get("index")

            # get the relation between the infill and the solution from the population
            rel = get_relation(infill, pop[k])

            # if the new solution is not dominated by the individual
            if rel <= 0:
                pool.append(infill)

            # if the individual is not dominated by the new solution
            if rel >= 0:
                pool.append(pop[k])

        # set the rank and crowding in the current population
        pool = Population.create(*pool)
        self.pop = RankAndCrowdingSurvival().do(self.problem, pool, n_survive=self.pop_size)


parse_doc_string(GDE3.__init__)



import numpy as np

from pymoo.core.algorithm import Algorithm
from pymoo.core.individual import Individual
from pymoo.core.meta import Meta
from pymoo.core.problem import Problem
from pymoo.operators.repair.to_bound import set_to_bounds_if_outside


class LineSearch(Algorithm):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.point, self.direction = None, None

    def setup(self, problem, point=None, direction=None, **kwargs):
        super().setup(problem, **kwargs)

        msg = "Only problems with one objective and no constraints can be solved using a line search!"
        assert not problem.has_constraints() and problem.n_obj == 1, msg

        assert point is not None, "You have to define a starting point for the algorithm"
        self.point = point

        assert direction is not None, "You have to define a direction point for the algorithm"
        self.direction = direction

        return self

    def _initialize_infill(self):

        # x could be a vector or an individual
        if isinstance(self.point, np.ndarray):
            self.point = Individual(X=self.point)

        # make sure it is evaluated - if not yet also get the gradient
        if self.point.get("F") is None:
            self.evaluator.eval(self.problem, self.point, algorithm=self)

        self.infill = self.point


class LineSearchProblem(Meta, Problem):

    def __init__(self, problem, point, direction, strict_bounds=True, xl=0.0, xu=np.inf):
        super().__init__(problem)
        self.n_var = 1
        self.xl, self.xu = np.array([xl]), np.array([xu])

        self.point = point
        self.direction = direction
        self.strict_bounds = strict_bounds

    def _evaluate(self, alpha, out, *args, **kwargs):
        out["alpha"] = alpha

        x = self.point.X + alpha * self.direction
        if self.strict_bounds:
            x = set_to_bounds_if_outside(x, self.xl, self.xu)
        out["__X__"] = x

        super()._evaluate(x, out, *args, **kwargs)



import autograd.numpy as anp
import numpy as np
from autograd import value_and_grad

from pymoo.util.normalization import normalize
from pymoo.util.ref_dirs.energy import squared_dist
from pymoo.util.ref_dirs.optimizer import Adam
from pymoo.util.reference_direction import ReferenceDirectionFactory, scale_reference_directions


class LayerwiseRieszEnergyReferenceDirectionFactory(ReferenceDirectionFactory):

    def __init__(self,
                 n_dim,
                 partitions,
                 return_as_tuple=False,
                 n_max_iter=1000,
                 verbose=False,
                 X=None,
                 **kwargs):

        super().__init__(n_dim, **kwargs)
        self.scalings = None
        self.n_max_iter = n_max_iter
        self.verbose = verbose
        self.return_as_tuple = return_as_tuple
        self.X = X
        self.partitions = partitions

    def _step(self, optimizer, X, scalings):
        obj, grad = value_and_grad(calc_potential_energy)(scalings, X)
        scalings = optimizer.next(scalings, np.array(grad))
        scalings = normalize(scalings, xl=0, xu=scalings.max())
        return scalings, obj

    def _solve(self, X, scalings):

        # initialize the optimizer for the run
        optimizer = Adam()

        # for each iteration of gradient descent
        for i in range(self.n_max_iter):

            # execute one optimization step
            _scalings, _obj = self._step(optimizer, X, scalings)

            # evaluate how much the points have moved
            delta = np.abs(_scalings - scalings).sum()

            if self.verbose:
                print(i, "objective", _obj, "delta", delta)

            # if there was only a little delta during the last iteration -> terminate
            if delta < 1e-5:
                scalings = _scalings
                break

            # otherwise use the new points for the next iteration
            scalings = _scalings

        self.scalings = scalings
        return get_points(X, scalings)

    def do(self):

        X = []
        scalings = []

        for k, p in enumerate(self.partitions):

            if p > 1:
                val = np.linspace(0, 1, p + 1)[1:-1]
                _X = []
                for i in range(self.n_dim):
                    for j in range(i + 1, self.n_dim):
                        x = np.zeros((len(val), self.n_dim))
                        x[:, i] = val
                        x[:, j] = 1 - val
                        _X.append(x)

                X.append(np.row_stack(_X + [np.eye(self.n_dim)]))

            elif p == 1:
                X.append(np.eye(self.n_dim))
            else:
                X.append(np.full(self.n_dim, 1 / self.n_dim)[None, :])

            scalings.append(1 - k / len(self.partitions))

        scalings = np.array(scalings)
        X = self._solve(X, scalings)

        return X


# ---------------------------------------------------------------------------------------------------------
# Energy Functions
# ---------------------------------------------------------------------------------------------------------


def get_points(X, scalings):
    vals = []
    for i in range(len(X)):
        vals.append(scale_reference_directions(X[i], scalings[i]))
    X = anp.row_stack(vals)
    return X


def calc_potential_energy(scalings, X):
    X = get_points(X, scalings)

    i, j = anp.triu_indices(len(X), 1)
    D = squared_dist(X, X)[i, j]

    if np.any(D < 1e-12):
        return np.nan, np.nan

    return (1 / D).mean()


import numpy as np

from pymoo.constraints.adaptive import AdaptiveConstraintHandling


class AdaptiveEpsilonConstraintHandling(AdaptiveConstraintHandling):

    def __init__(self, algorithm, perc_eps_until=0.5):
        super().__init__(algorithm)
        self.perc_eps_until = perc_eps_until
        self.max_cv = None

    def _adapt(self, config, **kwargs):
        t = self.termination.perc
        alpha = np.maximum(0.0, 1 - 1 / self.perc_eps_until * t)
        eps = alpha * self.max_cv

        config["cv_eps"] = eps

    def _initialize_advance(self, infills=None, **kwargs):

        # get the average constraint violation in the current generation
        cv = infills.get("cv")
        self.max_cv = np.mean(cv)

        return super()._initialize_advance(infills, **kwargs)



import numpy as np
from scipy.spatial.distance import cdist

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.algorithm import LoopwiseAlgorithm
from pymoo.core.duplicate import NoDuplicateElimination
from pymoo.core.population import Population
from pymoo.core.selection import Selection
from pymoo.core.variable import Real, get
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.reference_direction import default_ref_dirs


class NeighborhoodSelection(Selection):

    def __init__(self, prob=1.0) -> None:
        super().__init__()
        self.prob = Real(prob, bounds=(0.0, 1.0))

    def _do(self, problem, pop, n_select, n_parents, neighbors=None, **kwargs):
        assert n_select == len(neighbors)
        P = np.full((n_select, n_parents), -1)

        prob = get(self.prob, size=n_select)

        for k in range(n_select):
            if np.random.random() < prob[k]:
                P[k] = np.random.choice(neighbors[k], n_parents, replace=False)
            else:
                P[k] = np.random.permutation(len(pop))[:n_parents]

        return P


# =========================================================================================================
# Implementation
# =========================================================================================================

class MOEAD(LoopwiseAlgorithm, GeneticAlgorithm):

    def __init__(self,
                 ref_dirs=None,
                 n_neighbors=20,
                 decomposition=None,
                 prob_neighbor_mating=0.9,
                 sampling=FloatRandomSampling(),
                 crossover=SBX(prob=1.0, eta=20),
                 mutation=PM(prob_var=None, eta=20),
                 output=MultiObjectiveOutput(),
                 **kwargs):

        # reference directions used for MOEAD
        self.ref_dirs = ref_dirs

        # the decomposition metric used
        self.decomposition = decomposition

        # the number of neighbors considered during mating
        self.n_neighbors = n_neighbors

        self.neighbors = None

        self.selection = NeighborhoodSelection(prob=prob_neighbor_mating)

        super().__init__(pop_size=len(ref_dirs),
                         sampling=sampling,
                         crossover=crossover,
                         mutation=mutation,
                         eliminate_duplicates=NoDuplicateElimination(),
                         output=output,
                         advance_after_initialization=False,
                         **kwargs)

    def _setup(self, problem, **kwargs):
        assert not problem.has_constraints(), "This implementation of MOEAD does not support any constraints."

        # if no reference directions have been provided get them and override the population size and other settings
        if self.ref_dirs is None:
            self.ref_dirs = default_ref_dirs(problem.n_obj)
        self.pop_size = len(self.ref_dirs)

        # neighbours includes the entry by itself intentionally for the survival method
        self.neighbors = np.argsort(cdist(self.ref_dirs, self.ref_dirs), axis=1, kind='quicksort')[:, :self.n_neighbors]

        # if the decomposition is not set yet, set the default
        if self.decomposition is None:
            self.decomposition = default_decomp(problem)

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self.ideal = np.min(self.pop.get("F"), axis=0)

    def _next(self):
        pop = self.pop

        # iterate for each member of the population in random order
        for k in np.random.permutation(len(pop)):
            # get the parents using the neighborhood selection
            P = self.selection.do(self.problem, pop, 1, self.mating.crossover.n_parents, neighbors=[self.neighbors[k]])

            # perform a mating using the default operators - if more than one offspring just pick the first
            off = np.random.choice(self.mating.do(self.problem, pop, 1, parents=P, n_max_iterations=1))

            # evaluate the offspring
            off = yield off

            # update the ideal point
            self.ideal = np.min(np.vstack([self.ideal, off.F]), axis=0)

            # now actually do the replacement of the individual is better
            self._replace(k, off)

    def _replace(self, k, off):
        pop = self.pop

        # calculate the decomposed values for each neighbor
        N = self.neighbors[k]
        FV = self.decomposition.do(pop[N].get("F"), weights=self.ref_dirs[N, :], ideal_point=self.ideal)
        off_FV = self.decomposition.do(off.F[None, :], weights=self.ref_dirs[N, :], ideal_point=self.ideal)

        # this makes the algorithm to support constraints - not originally proposed though and not tested enough
        # if self.problem.has_constraints():
        #     CV, off_CV = pop[N].get("CV")[:, 0], np.full(len(off_FV), off.CV)
        #     fmax = max(FV.max(), off_FV.max())
        #     FV, off_FV = parameter_less(FV, CV, fmax=fmax), parameter_less(off_FV, off_CV, fmax=fmax)

        # get the absolute index in F where offspring is better than the current F (decomposed space)
        I = np.where(off_FV < FV)[0]
        pop[N[I]] = off


class ParallelMOEAD(MOEAD):

    def __init__(self, ref_dirs, **kwargs):
        super().__init__(ref_dirs, **kwargs)
        self.indices = None

    def _infill(self):
        pop_size, cross_parents, cross_off = self.pop_size, self.mating.crossover.n_parents, self.mating.crossover.n_offsprings

        # do the mating in a random order
        indices = np.random.permutation(len(self.pop))[:self.n_offsprings]

        # get the parents using the neighborhood selection
        P = self.selection.do(self.problem, self.pop, self.n_offsprings, cross_parents,
                              neighbors=self.neighbors[indices])

        # do not any duplicates elimination - thus this results in exactly pop_size * n_offsprings offsprings
        off = self.mating.do(self.problem, self.pop, 1e12, n_max_iterations=1, parents=P)

        # select a random offspring from each mating
        off = Population.create(*[np.random.choice(pool) for pool in np.reshape(off, (self.n_offsprings, -1))])

        # store the indices because of the neighborhood matching in advance
        self.indices = indices

        return off

    def _advance(self, infills=None, **kwargs):
        assert len(self.indices) == len(infills), "Number of infills must be equal to the one created beforehand."

        # update the ideal point before starting to replace
        self.ideal = np.min(np.vstack([self.ideal, infills.get("F")]), axis=0)

        # now do the replacements as in the loop-wise version
        for k, off in enumerate(infills):
            self._replace(self.indices[k], off)


def default_decomp(problem):
    if problem.n_obj <= 2:
        from pymoo.decomposition.tchebicheff import Tchebicheff
        return Tchebicheff()
    else:
        from pymoo.decomposition.pbi import PBI
        return PBI()


parse_doc_string(MOEAD.__init__)
